# $Id:$
# -*- coding: latin-1 -*-
## vendredi 3 juillet 2009, 20:10:54 (UTC+0200)
## ********************************************************************************
## compute.py dim jan 23 20:39:02 CET 2005
## Copyright (C) 2004-2009 Grard Deneux
## gerard.deneux@free.fr

## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.

## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
## Lesser General Public License for more details.
## http://www.opensource.org/licenses/lgpl-license.php

## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
## ******************************************************************************
try:
    import math, cmath, locale
    from types import *
except ImportError, e:
    print e
#=========================================
# a class usefull for various computations
#=========================================
class Compute():
    """Utile pour divers calculs
    Usefull for various computations"""
    def __init__(self, ctx):
        self.ctx = ctx
    # **********************************
    # show message wrapper
    # **********************************
    def showDebugMsg(self, aMsg):
        if not hasattr(self, "xMessage"):
            from fr.free.deneux.gerard.cadooo.messagebox import MessageBox
            self.xMessage = MessageBox(self.ctx)
        self.xMessage.execute(aMsg, "infobox")
        return
    # *************************************************
    # transform an area measure from system unit to user unit
    # *************************************************
    def transform2UserSqUnit(self, fMeasure, sUnits, fUnitsRatio, nUnitIdx, fDrawScale):
        fUserMeasure = fMeasure * (fUnitsRatio[nUnitIdx] / fDrawScale)**2
        nPrecision = int(abs(math.log10((fUnitsRatio[nUnitIdx] / fDrawScale)**2)))
        return "%.*f %s 2" % (nPrecision, fUserMeasure, sUnits[nUnitIdx])
    # *************************************************
    # transform an area measure from system unit to user unit
    # *************************************************
    def transform2UserUnitAsFloat(self, fMeasure, sUnits, fUnitsRatio, nUnitIdx, fDrawScale):
        fUserMeasure = fMeasure * (fUnitsRatio[nUnitIdx] / fDrawScale)
        return fUserMeasure
    # *************************************************
    # transform a measure from system unit to user unit
    # *************************************************
    def transform2UserUnit(self, fMeasure, sUnits, fUnitsRatio, nUnitIdx, fDrawScale):
        loc = locale.setlocale(locale.LC_ALL)
        locale.setlocale(locale.LC_ALL, '')
        fUserMeasure = fMeasure * fUnitsRatio[nUnitIdx] / fDrawScale
        # ------------------------------------------
        if sUnits[nUnitIdx] in ('mm', 'pt', 'pica'):
            sMeasure = locale.format("%.1f", round(fUserMeasure, 1))
        # ----------------------------
        elif sUnits[nUnitIdx] == 'cm':
            sMeasure = locale.format("%.2f", round(fUserMeasure, 2))
        # ---------------------------
        elif sUnits[nUnitIdx] == 'm':
            sMeasure = locale.format("%.4f", round(fUserMeasure, 4))
        # ----------------------------
        elif sUnits[nUnitIdx] == 'km':
            sMeasure = locale.format("%.7f", round(fUserMeasure, 7))
        # ------------------------------
        elif sUnits[nUnitIdx] == 'inch':
            nInch = long(fUserMeasure)
            nRem = long(round((fUserMeasure - nInch) * 32))
            nDiv = 1
            while nRem % 2 == 0 and nRem > 1:
                nRem /= 2
                nDiv *=2
            sMeasure = "%d\" %d/%d" % (nInch, nRem, 32/nDiv)
        # ------------------------------
        elif sUnits[nUnitIdx] == 'feet':
            nFeet = long(fUserMeasure)
            nInch = long(round((fUserMeasure - nFeet) * 12))
            nRem = long(round(((fUserMeasure - nFeet) * 12 - nInch) * 32))
            nDiv = 1
            while nRem % 2 == 0 and nRem > 1:
                nRem /= 2
                nDiv *=2
            sMeasure = "%d\' %d\" %d/%d" % (nFeet, nInch, nRem, 32/nDiv)
        # -------------------------------
        elif sUnits[nUnitIdx] == 'miles':
            nMi = long(fUserMeasure)
            nYrd = long(round((fUserMeasure - nMi) * 1760))
            nFeet = long(round(((fUserMeasure - nMi) * 1760 - nYrd) * 3))
            nInch = long(round((((fUserMeasure - nMi) * 1760 - nYrd) * 3 - nFeet) * 12))
            nRem = long(round((((fUserMeasure - nMi) * 1760 - nYrd) * 3 - nFeet) * 12 - nInch) * 32)
            nDiv = 1
            while nRem % 2 == 0 and nRem > 1:
                nRem /= 2
                nDiv *=2
            sMeasure = "%d mi %d yrd %d\' %d\" %d/%d" % (nMi, nYrd, nInch, nRem, 32/nDiv)
        locale.setlocale(locale.LC_ALL, loc)
        return sMeasure
    # *************************************************
    # transform a measure from user unit to system unit
    # *************************************************
    def transform2SystemUnit(self, sMeasure, sUnits, fUnitsRatio, nUnitIdx, fDrawScale):
        loc = locale.setlocale(locale.LC_ALL)
        locale.setlocale(locale.LC_ALL, '')
        # ----------------------------------------------------
        if sUnits[nUnitIdx] in ('mm', 'cm', 'm', 'km', 'pt', 'pica'):
            fUserMeasure = locale.atof(sMeasure.decode('string_escape'))
        # ------------------------------------------
        elif sUnits[nUnitIdx] in ('inch', 'feet', 'miles'):
            fMi = fYrd = fFeet = fInch = fRem = 0.0
            # ......................
            if sUnits[nUnitIdx].count('mi'):
                fMi = float(sMeasure.split('mi')[0])
                sMeasure = sMeasure.split('mi')[1]
            # .......................
            if sMeasure.count('yrd'):
                fYrd = float(sMeasure.split('yrd')[0])
                sMeasure = sMeasure.split('yrd')[1]
            # .....................
            if sMeasure.count("'"):
                fFeet = float(sMeasure.split("'")[0])
                sMeasure = sMeasure.split("'")[1]
            # .....................
            if sMeasure.count('"'):
                fInch = float(sMeasure.split('"')[0])
                sMeasure = sMeasure.split('"')[1]
            # .....................
            if sMeasure.count('/'):
                fRem = float(sMeasure.split('/')[0]) / sMeasure.split('/')[1]
            # ............................
            if sUnits[nUnitIdx] == "inch":
                fUserMeasure = (((fMi * 1760 + fYrd) * 3) + fFeet) * 12 + fInch + fRem
            # ............................
            if sUnits[nUnitIdx] == "feet":
                fUserMeasure = ((fMi * 1760 + fYrd) * 3) + fFeet + (fInch + fRem) / 12
            # ............................
            if sUnits[nUnitIdx] == "miles":
                fUserMeasure = fMi + (fYrd + (fFeet + (fInch + fRem) / 12) / 3) / 1760 
        locale.setlocale(locale.LC_ALL, loc)
        return long(round(fUserMeasure / fUnitsRatio[nUnitIdx] * fDrawScale))
    # ************************************
    # transform cordinates from UCS to GCS
    # ************************************
    def transform2GCS(self, aUCSCoord, aUcs):
        aGCSCoord = {'X': 0, 'Y': 0}
        aGCSCoord['X'] = aUcs[0][2] + aUCSCoord['X'] * aUcs[0][0] + aUCSCoord['Y'] * aUcs[0][1]
        aGCSCoord['Y'] = aUcs[1][2] + aUCSCoord['X'] * aUcs[1][0] + aUCSCoord['Y'] * aUcs[1][1]
        return aGCSCoord
    # ************************************
    # transform cordinates from GCS to UCS
    # ************************************
    def transform2UCS(self, aGCSCoord, aUcs):
        aUCSCoord = {'X': 0, 'Y': 0}
        aUCSCoord['X'] = (aGCSCoord['X'] - aUcs[0][2]) * aUcs[0][0] + (aGCSCoord['Y'] - aUcs[1][2]) * aUcs[1][0]
        aUCSCoord['Y'] = (aGCSCoord['X'] - aUcs[0][2]) * aUcs[0][1] + (aGCSCoord['Y'] - aUcs[1][2]) * aUcs[1][1]
        return aUCSCoord
    # *********************************************
    # transform cordinates from GCS to UCS and unit
    # *********************************************
    def transform2UCSAndUnit(self, aGCSCoord, aEnv):
        aCoord = self.transform2UCS(aGCSCoord, aEnv['UCS'])
        aCoord['X'] = self.transform2UserUnit(aCoord['X'], aEnv['UNITS'], aEnv['UNITSRATIO'], aEnv['MEASUREUNIT'], aEnv['DRAWSCALE'])
        aCoord['Y'] = self.transform2UserUnit(aCoord['Y'], aEnv['UNITS'], aEnv['UNITSRATIO'], aEnv['MEASUREUNIT'], aEnv['DRAWSCALE'])
        return aCoord
    # ******************************************************
    # transform cordinates from GCS to UCS and unit as float
    # ******************************************************
    def transform2UCSAndUnitAsFloat(self, aGCSCoord, aEnv):
        aCoord = self.transform2UCS(aGCSCoord, aEnv['UCS'])
        aCoord['X'] = self.transform2UserUnitAsFloat(aCoord['X'], aEnv['UNITS'], aEnv['UNITSRATIO'], aEnv['MEASUREUNIT'], aEnv['DRAWSCALE'])
        aCoord['Y'] = self.transform2UserUnitAsFloat(aCoord['Y'], aEnv['UNITS'], aEnv['UNITSRATIO'], aEnv['MEASUREUNIT'], aEnv['DRAWSCALE'])
        return aCoord
    # ***************************************************
    # transform screen coordinates to drawing coordinates
    # ***************************************************
    def transformMouseCoord(self, aEnv):
        aMouseCoord = {"X": 0,"Y": 0}
        aMouseCoord["X"] = long(round(float(aEnv['MOUSEEVENT'].X) * aEnv['DEVICEXSCALE'] + aEnv['VISIBLEAREA']['X']))
        aMouseCoord["Y"] = long(round(float(aEnv['MOUSEEVENT'].Y) * aEnv['DEVICEYSCALE'] + aEnv['VISIBLEAREA']['Y']))
        if aEnv['ORTHOMODE'] and aEnv['VALIDORIGIN']:
            if aMouseCoord["X"] < aEnv['ORIGINCOORD']['X']:
                nSignX = -1
            else:
                nSignX = 1
            if aMouseCoord["Y"] < aEnv['ORIGINCOORD']['Y']:
                nSignY = -1
            else:
                nSignY = 1
            if abs(aMouseCoord["Y"] - aEnv['ORIGINCOORD']['Y']) < (math.sqrt(2)-1)*abs(aMouseCoord["X"] - aEnv['ORIGINCOORD']['X']):
                aMouseCoord["Y"] = aEnv['ORIGINCOORD']['Y']
            elif abs(aMouseCoord["Y"] - aEnv['ORIGINCOORD']['Y']) < abs(aMouseCoord["X"] - aEnv['ORIGINCOORD']['X']):
                aMouseCoord["Y"] = aEnv['ORIGINCOORD']['Y'] + nSignY * abs(aMouseCoord["X"] - aEnv['ORIGINCOORD']['X'])
            elif abs(aMouseCoord["Y"] - aEnv['ORIGINCOORD']['Y']) < (math.sqrt(2)+1)*abs(aMouseCoord["X"] - aEnv['ORIGINCOORD']['X']):
                aMouseCoord["X"] = aEnv['ORIGINCOORD']['X'] + nSignX * abs(aMouseCoord["Y"] - aEnv['ORIGINCOORD']['Y'])
            else:
                aMouseCoord["X"] = aEnv['ORIGINCOORD']['X'] + 1
        return aMouseCoord
    ##
    ## Convert rectangular coordinates to polar - conversion coordonnes rectangulaires en polaires
    ##
    def toPolarCoord(self, aCoord0, aCoord1):
        """Conversion de coordonnees rectangulaires en polaires
        Polar to rectangular coordinates conversion
        @param aCoord0 as com.sun.star.awt.Point
        @param aCoord1 as com.sun.star.awt.Point
        return (distance, angle) as tuple(int, float)"""
        aPolarCoord = []
        aPolarCoord.append(long(round(math.hypot(float(aCoord1["X"] - aCoord0["X"]), (aCoord1["Y"] - aCoord0["Y"])))))
        if aPolarCoord[0] == 0:
            aPolarCoord.append(0)
            return tuple(aPolarCoord)
        aPolarCoord.append(math.asin(float(abs(aCoord1["Y"] - aCoord0["Y"])) / aPolarCoord[0]))
        if (aCoord1["X"] - aCoord0["X"]) >= 0:
            if (aCoord1["Y"] - aCoord0["Y"]) < 0:
                aPolarCoord[1] = 2 * math.pi - aPolarCoord[1]
        else:
            if (aCoord1["Y"] - aCoord0["Y"]) > 0:
                aPolarCoord[1] = math.pi - aPolarCoord[1]
            else:
                aPolarCoord[1] = math.pi + aPolarCoord[1]
        if aPolarCoord[1] > 0:
            aPolarCoord[1] = 2* math.pi - aPolarCoord[1]
        return tuple(aPolarCoord)
    ##
    ## Compute balanced steps - calcul du balancement
    ##
    def getCollarX(self, nCollarLength, nStrideLength, nStrideX):
        """Calcul du balancement sur la ligne de collet
        balance computing on collar line
        @param nCollarLength as int
        @param nStrideLength as int
        @param nStrideX as int
        return nCollarX as int"""
        fKg = float(nStrideX) / nStrideLength
        fK = float(nCollarLength) / nStrideLength
        if nStrideX > long(round((nStrideLength - nCollarLength) / float(nStrideLength + nCollarLength) * nCollarLength)):
            return long(round(nStrideX * ((fK * (1 + fK ** 2)) / (2 * fK ** 2 + (1 - fK ** 2) * fKg))))
        else:
            return nStrideX
    ##
    ## Compute the coordinates of two lines intersection point and the distances from each parameter point coordinates
    ## Calcule les distances entre les points donnes en parametre et le point d'intersection des droites
    ##
    def getInterLines3(self, aCoordA, fAngle1, aCoordB, fAngle2, nOffset1, nOffset2):
        """Calcul des coordonnees du point d'intersection de deux droites et des distances par rapport aux points donnes en parametre
        Compute the coordinates of two lines intersection point and the distances from each parameter point coordinates
        @param aCoordA, aCoordC as com.sun.star.awt.Point
        @param fAngle1, fAngle2 as float
        @param nOffset1, nOffset2 as Long
        return nDistance1, nDistance2, nX, nY as tuple(int, int, int, int)"""
        aCoord1 = {"X": 0,"Y": 0}
        aCoord2 = {"X": 0,"Y": 0}
        aCoord1["X"] = aCoordA["X"] + long(round(nOffset1 * math.sin(fAngle1)))
        aCoord1["Y"] = aCoordA["Y"] + long(round(nOffset1 * math.cos(fAngle1)))
        aCoord2["X"] = aCoordB["X"] + long(round(nOffset2 * math.sin(fAngle2)))
        aCoord2["Y"] = aCoordB["Y"] + long(round(nOffset2 * math.cos(fAngle2)))
        return self.getInterLines(aCoord1, fAngle1, aCoord2, fAngle2)
    ##
    ## Compute the coordinates of two lines intersection point and the distances from each parameter point coordinates
    ## Calcule les distances entre les points donnes en parametre et le point d'intersection des droites
    ##
    def getInterLines2(self, aCoordA, aCoordB, aCoordC, aCoordD, nOffset1, nOffset2):
        """Calcul des coordonnees du point d'intersection de deux droites et des distances par rapport aux points donnes en parametre
        Compute the coordinates of two lines intersection point and the distances from each parameter point coordinates
        @param aCoordA, aCoordB, aCoordC, aCoordD as com.sun.star.awt.Point
        @param nOffset1, nOffset2 as Long
        return nDistance1, nDistance2, nX, nY as tuple(int, int, int, int)"""
        aCoord1 = self.getCoordFromAbsOrd(aCoordA, aCoordB, 0, nOffset1)
        aCoord2 = self.getCoordFromAbsOrd(aCoordC, aCoordD, 0, nOffset2)
        fAngle1 = self.toPolarCoord(aCoordA, aCoordB)[1]
        fAngle2 = self.toPolarCoord(aCoordC, aCoordD)[1]
        return self.getInterLines(aCoord1, fAngle1, aCoord2, fAngle2)
    ##
    ## Compute the distances from each parameter point coordinates to the intersection point of the two lines
    ## Calcule les distances entre les points donnes en parametre et le point d'intersection des droites
    ##
    def getInterLines(self, aCoord1, fAngle1, aCoord2, fAngle2):
        """Calcul des coordonnees du point d'intersection de deux droites et des distances par rapport aux points donnes en parametre
        Compute the coordinates of two lines intersection point and the distances from each parameter point coordinates
        @param aCoord1 as com.sun.star.awt.Point
        @param fAngle1 as float
        @param aCoord2 as com.sun.star.awt.Point
        @param fAngle2 as float
        return nDistance1, nDistance2, nX, nY as tuple(int, int, int, int)"""
        nInter = [0 for i in range(4)]
        if abs(math.cos(fAngle1)) < 1e-3:
            nInter[2] = aCoord1["X"]
            if abs(math.cos(fAngle2)) >= 1e-3:
                nInter[3] = aCoord2["Y"] - long(round(math.tan(fAngle2)*(aCoord1["X"] - aCoord2["X"])))
                nInter[0] = long(round((nInter[3] - aCoord1["Y"]) / math.sin(-fAngle1)))
                nInter[1] = long(round((nInter[2] - aCoord2["X"]) / math.cos(fAngle2)))
                return tuple(nInter)
            else:
                return None
        elif abs(math.cos(fAngle2)) < 1e-3:
            nInter[2] = aCoord2["X"]
            nInter[3] = aCoord1["Y"] - long(round(math.tan(fAngle1)*(aCoord2["X"] - aCoord1["X"])))
            nInter[1] = long(round((nInter[3] - aCoord2["Y"]) / math.sin(-fAngle2)))
            nInter[0] = long(round((nInter[2] - aCoord1["X"]) / math.cos(fAngle1)))
            return tuple(nInter)
        else:
            fM1 = -math.tan(fAngle1)
            fM2 = -math.tan(fAngle2)
            if abs(fM1 - fM2) < 1e-3:
                return None
            nP1 = aCoord1["Y"] - long(round(fM1 * aCoord1["X"]))
            nP2 = aCoord2["Y"] - long(round(fM2 * aCoord2["X"]))
            nInter[2] = long(round(-(nP1 - nP2) / (fM1 - fM2)))
            nInter[3] = long(round(fM1 * nInter[2])) + nP1
            nInter[0] = long(round((nInter[2] - aCoord1["X"]) / math.cos(fAngle1)))
            nInter[1] = long(round((nInter[2] - aCoord2["X"]) / math.cos(fAngle2)))
        return tuple(nInter)
    ##
    ## Compute the distances from first parameter point coordinates to each intersection point with the ellipse
    ## Calcule les distances entre le point de la droite donne en parametre et les points d'intersection avec l'ellipse
    ##
    def getInterLineEllipse2(self, aCoordA, aCoordB, aCoordC, aSize, fRotation):
        """Compute distances and coordinates for each intersection point between a line and an ellipse
        Calcule distances et coordonnes des points d'intersection d'une droite avec une ellipse
        @param aCoordA as com.sun.star.awt.Point
        @param aCoordB as com.sun.star.awt.Point
        @param aCoordC as com.sun.star.awt.Point
        @param aSize as com.sun.star.awt.Size
        @param fRotation as float
        return nDistance1, nDistance2, nX1, nY1, nX2, nY2 as tuple(long, long, long, long, long, long)"""
        fAngle = self.toPolarCoord(aCoordA, aCoordB)[1]
        return self.getInterLineEllipse(aCoordA, fAngle, aCoordC, aSize, fRotation)
    ##
    ## Compute the distances from first parameter point coordinates to each intersection point with the ellipse
    ## Calcule les distances entre le point de la droite donne en parametre et les points d'intersection avec l'ellipse
    ##
    def getInterLineEllipse(self, aCoord1, fAngle1, aCoord2, aSize, fAngle2):
        """Compute distances and coordinates for each intersection point between a line and an ellipse
        Calcule distances et coordonnes des points d'intersection d'une droite avec une ellipse
        @param aCoord1 as com.sun.star.awt.Point
        @param fAngle1 as float
        @param aCoord2 as com.sun.star.awt.Point
        @param aSize as com.sun.star.awt.Size
        @param fAngle2 as float
        return nDistance1, nDistance2, nX1, nY1, nX2, nY2 as tuple(int, int, int, int, int, int)"""
        nInter = []
        fAlfa1 = 2 * math.pi - fAngle1
        fAlfa2 = 2 * math.pi - fAngle2
        fRotation = fAlfa1 - fAlfa2
        aCoord = {'X': 0, 'Y': 0}
        aCoord["X"] = long(round((aCoord1["X"]-aCoord2["X"])*math.cos(fAlfa2)+(aCoord1["Y"]-aCoord2["Y"])*math.sin(fAlfa2)))
        aCoord["Y"] = long(round(-(aCoord1["X"]-aCoord2["X"])*math.sin(fAlfa2)+(aCoord1["Y"]-aCoord2["Y"])*math.cos(fAlfa2)))
        if abs(math.pi/2 - abs(fRotation)) < 1e-6 or abs(3*math.pi/2 - abs(fRotation)) < 1e-6:
            if abs(aCoord["X"]) > aSize["Width"]:
                return None
            else:
                nInter.append(long(round((-aSize["Height"]*math.sqrt(1-(float(aCoord["X"])/aSize["Width"])**2)-aCoord["Y"])/math.sin(fRotation))))
                nInter.append(long(round((aSize["Height"]*math.sqrt(1-(float(aCoord["X"])/aSize["Width"])**2)-aCoord["Y"])/math.sin(fRotation))))
        elif abs(fRotation) < 1e-6 or abs(math.pi - abs(fRotation)) < 1e-6:
            if abs(aCoord["Y"]) > aSize["Height"]:
                return None
            else:
                nInter.append(-long(round((aSize["Width"]*math.sqrt(1-(float(aCoord["Y"])/aSize["Height"])**2)-aCoord["X"])/math.cos(fRotation))))
                nInter.append(long(round((aSize["Width"]*math.sqrt(1-(float(aCoord["Y"])/aSize["Height"])**2)-aCoord["X"])/math.cos(fRotation))))
        else:
            fM = math.tan(fRotation)
            fP = aCoord["Y"] - fM * aCoord["X"]
            fA = (aSize["Width"] * fM)**2 + aSize["Height"]**2
            fDelta = fA - fP**2
            if fDelta < 0:
                return None
            else:
                nInter.append(long(round(((aSize["Width"]*(aSize["Height"]*math.sqrt(fDelta)-aSize["Width"]*fM*fP)/fA)-aCoord["X"])/math.cos(fRotation))))
                nInter.append(long(round(((-aSize["Width"]*(aSize["Height"]*math.sqrt(fDelta)+aSize["Width"]*fM*fP)/fA)-aCoord["X"])/math.cos(fRotation))))
        nInter.append(aCoord1["X"] + long(round(nInter[0] * math.cos(fAlfa1))))
        nInter.append(aCoord1["Y"] + long(round(nInter[0] * math.sin(fAlfa1))))
        nInter.append(aCoord1["X"] + long(round(nInter[1] * math.cos(fAlfa1))))
        nInter.append(aCoord1["Y"] + long(round(nInter[1] * math.sin(fAlfa1))))
        return tuple(nInter)
    ##
    ## Compute coordinates for each intersection point between 2 ellipses
    ## Calcule les coordonnes des points d'intersection de deux ellipses
    ##
    def getInterEllipses(self, aCoord0, aSize0, fAngle0, aCoord1, aSize1, fAngle1):
        """Compute coordinates for each intersection point between 2 ellipses
        Calcule coordonnes des points d'intersection de deux ellipses
        @param aCoord0 as com.sun.star.awt.Point : coordinates of first ellipse center
        @param aSize0 as com.sun.star.awt.Size : half short and great axis dimension
        @param fAngle0 as float : rotation angle
        @param aCoord1 as com.sun.star.awt.Point : coordinates of second ellipse center
        @param aSize0 as com.sun.star.awt.Size : half short and great axis dimension
        @param fAngle1 as float : rotation angle
        return nX1, nY1, nX2, nY2, nX3, nY3, nX4, nY4 as tuple(int, int, int, int, int, int, int, int)"""
        if abs(aCoord0["X"]-aCoord1["X"])<5 and abs(aCoord0["Y"]-aCoord1["Y"])<5 and (aSize0["Width"]-aSize1["Width"])*(aSize0["Height"]-aSize1["Height"])>=0 and abs(fAngle0-fAngle1)<1e-6:
            return None
        nInterCoord = []
        if abs(aSize0["Width"]-aSize0["Height"])<10 and abs(aSize1["Width"]-aSize1["Height"])<10:
            nDist, fTheta0 = self.toPolarCoord(aCoord0, aCoord1)
            if nDist == 0:
                return None
            if aSize0["Width"]+aSize1["Width"]<nDist:
                return None
            nU = (aSize0["Width"]**2+nDist**2-aSize1["Width"]**2)/(2*nDist)
            if abs(nU) > aSize0["Width"]:
                return None
            nV = long(round(math.sqrt(aSize0["Width"]**2-nU**2)))
            nInterCoord.append(long(round(aCoord0["X"] + nU * math.cos(fTheta0) - nV * math.sin(-fTheta0))))
            nInterCoord.append(long(round(aCoord0["Y"] + nU * math.sin(-fTheta0) + nV * math.cos(fTheta0))))
            nInterCoord.append(long(round(aCoord0["X"] + nU * math.cos(fTheta0) + nV * math.sin(-fTheta0))))
            nInterCoord.append(long(round(aCoord0["Y"] + nU * math.sin(-fTheta0) - nV * math.cos(fTheta0))))
            return tuple(nInterCoord)
        fK = math.pow(10, -max(abs(round(math.log10(aSize0["Width"]))), abs(round(math.log10(aSize0["Height"]))), abs(round(math.log10(aSize1["Width"]))), abs(round(math.log10(aSize1["Height"])))))
        # First ellipse
        fC012 = float(aSize0["Width"] * fK)**2 - float(aSize0["Height"] * fK)**2
        fA000 = float(aSize0["Width"] * fK)**2 - fC012 * (math.cos(fAngle0))**2
        fA001 = -fC012 * math.sin(-2.0 * fAngle0)
        fA011 = float(aSize0["Width"] * fK)**2 - fC012 * (math.sin(fAngle0))**2
        fB00 = -2.0 * fA000 * aCoord0["X"] * fK - fA001 * aCoord0["Y"] * fK
        fB01 = -fA001 * aCoord0["X"] * fK - 2.0 * fA011 * aCoord0["Y"] * fK
        fC0 = -(aSize0["Width"] * aSize0["Height"] * fK**2)**2 + fA000 * (aCoord0["X"] * fK)**2 + fA011 * (aCoord0["Y"] * fK)**2 + fA001 * aCoord0["X"] * aCoord0["Y"] * fK**2
        # 2nd ellipse
        fC112 = float(aSize1["Width"] * fK)**2 - float(aSize1["Height"] * fK)**2
        # ------------------------------------------------------
        fA100 = float(aSize1["Width"] * fK)**2 - fC112 * (math.cos(fAngle1))**2
        fA101 = -fC112 * math.sin(-2.0 * fAngle1)
        fA111 = float(aSize1["Width"] * fK)**2 - fC112 * (math.sin(fAngle1))**2
        fB10 = -2.0 * fA100 * aCoord1["X"] * fK - fA101 * aCoord1["Y"] * fK
        fB11 = -fA101 * aCoord1["X"] * fK - 2.0 * fA111 * aCoord1["Y"] * fK
        fC1 = -(aSize1["Width"] * aSize1["Height"] * fK**2)**2 + fA100 * (aCoord1["X"] * fK)**2 + fA111 * (aCoord1["Y"] * fK)**2 + fA101 * aCoord1["X"] * aCoord1["Y"] * fK**2
        # -------------------------------------
        fV = []
        fV.append(fC0 * fB10 - fB00 * fC1)
        fV.append(fC0 * fA101 - fA001 * fC1)
        fV.append(fB01 * fB10 - fB00 * fB11)
        fV.append(fB01 * fA101 - fA001 * fB11)
        fV.append(fA011 * fB10 - fB00 * fA111)
        fV.append(fA011 * fA101 - fA001 * fA111)
        fV.append(fB00 * fA100 - fA000 * fB10)
        fV.append(fA001 * fA100 - fA000 * fA101)
        fV.append(fC0 * fA100 - fA000 * fC1)
        fV.append(fB01 * fA100 - fA000 * fB11)
        fV.append(fA011 * fA100 - fA000 * fA111)
        # -----------------------------------
        fCoeff = []
        fCoeff.append(fV[0]*fV[6] - fV[8]**2)
        fCoeff.append(fV[0]*fV[7] + fV[6]*(fV[1] + fV[2]) - 2.0*fV[8]*fV[9])
        fCoeff.append(fV[7]*(fV[1] + fV[2]) + fV[6]*(fV[3] + fV[4]) - fV[9]**2 - 2.0*fV[8]*fV[10])
        fCoeff.append(fV[7]*(fV[3] + fV[4]) + fV[5]*fV[6] - 2.0*fV[9]*fV[10])
        fCoeff.append(fV[5]*fV[7] - fV[10]**2)
        fInterCoordY = self.solvePoly4(fCoeff)
        nInterCoord = []
        if fInterCoordY is None:
            return None
        for fY in fInterCoordY:
            fAlfa = [fA011*fY**2 + fB01*fY + fC0, fA001*fY + fB00, fA000]
            fBeta = [fA111*fY**2 + fB11*fY + fC1, fA101*fY + fB10, fA100]
            fX = (fAlfa[2] * fBeta[0] - fAlfa[0] * fBeta[2])/(fAlfa[1] * fBeta[2] - fAlfa[2] * fBeta[1])
            nInterCoord.extend([long(round(fX/fK)), long(round(fY/fK))])
        return tuple(nInterCoord)
    ##
    ## Compute coordinates of the tangent point between a line  and circle
    ## Calcule les coordonnes du point de tangence entre une droite et un cercle
    ##
    def getTangentLineCircle2(self, aCoordA, aCoordB, aCoordC):
        """Compute coordinates of the tangent point between a line  and circle
        Calcule les coordonnes du point de tangence entre une droite et un cercle
        @param aCoordA as com.sun.star.awt.Point : coordinates of a point of the line
        @param aCoordB as com.sun.star.awt.Point : coordinates of a point of the line
        @param aCoordC as com.sun.star.awt.Point : coordinates of circle center
        return nX1, nY1 as tuple(long, long)"""
        fAngle = self.toPolarCoord(aCoordA, aCoordB)
        return self.getTangentLineCircle(aCoordA, fAngle, aCoordC)
    ##
    ## Compute coordinates of the tangent point between a line  and circle
    ## Calcule les coordonnes du point de tangence entre une droite et un cercle
    ##
    def getTangentLineCircle(self, aCoord1, fAngle1, aCoord2):
        """Compute coordinates of the tangent point between a line  and circle
        Calcule les coordonnes du point de tangence entre une droite et un cercle
        @param aCoord1 as com.sun.star.awt.Point : coordinates of a point of the line
        @param fAngle1 as float : slope angle of the line
        @param aCoord2 as com.sun.star.awt.Point : coordinates of circle center
        return nX1, nY1 as tuple(int, int)"""
        nTanCoord = []
        if math.sin(abs(fAngle1)) < 1e-3 :
            nTanCoord.append(aCoord2["X"])
            nTanCoord.append(aCoord1["Y"])
        elif math.cos(fAngle1) < 1e-3 :
            nTanCoord.append(aCoord1["X"])
            nTanCoord.append(aCoord2["Y"])
        else:
            fP = aCoord1["Y"] - aCoord2["Y"] - math.tan(-fAngle1) * (aCoord1["X"] - aCoord2["X"])
            if abs(fP) < 1e-3:
                nTanCoord.append(aCoord2["X"])
                nTanCoord.append(aCoord2["Y"])
            else:
                nTanCoord.append(long(round(aCoord2["X"] - math.tan(-fAngle1) * fP / (1 + math.tan(-fAngle1)**2))))
                nTanCoord.append(long(round(aCoord2["Y"] + math.tan(-fAngle1) * (nTanCoord[0] - aCoord2["X"]) + fP)))
        return tuple(nTanCoord)
    ##
    ## Solve degree 3 polynomial equation
    ## Rsolution d'une quation polynomiale de degrs 3
    ##
    def solvePoly3(self, fP3):
        """Solve polynomial equation of degree 3
        Rsolution d'une quation polynomiale de degrs 3
        @param fCoeff as tuple of float with length = 3 (degree 3 coeff being always 1)
        return zeros as tuple of float or None"""
        fZeros3 = []
        # -------------------------------------
        fP32 = []
        fP32.append(2.0 * fP3[2]**3 / 27.0 - fP3[2] * fP3[1] / 3.0 + fP3[0]) # h
        fP32.append(-fP3[2]**2 / 3.0 + fP3[1]) # g
        # -------------------------------------
        fP2 = []
        fP2.append(-fP32[1]**3 / 27.0)
        fP2.append(fP32[0])
        fZeros2 = self.solvePoly2(fP2)
        fJ = complex(math.cos(2 * math.pi / 3), math.sin(2 * math.pi / 3))
        if type(fZeros2[0]) is ComplexType:
            # discriminant ngatif pour la rsolution de l'quation de degrs 3
            fTheta = math.atan2(fZeros2[0].imag, fZeros2[0].real)
            fRho = abs(fZeros2[0])
            fU = complex(fRho**(1.0/3.0) * math.cos(fTheta / 3.0), fRho**(1.0/3.0) * math.sin(fTheta / 3.0))
            fTheta = math.atan2(fZeros2[1].imag, fZeros2[1].real)
            fRho = abs(fZeros2[1])
            fV = complex(fRho**(1.0/3.0) * math.cos(fTheta / 3.0), fRho**(1.0/3.0) * math.sin(fTheta / 3.0))
        else:
            fU = complex((1-2*(fZeros2[0]<0))*abs(fZeros2[0])**(1/3.0), 0.0)
            fV = complex((1-2*(fZeros2[1]<0))*abs(fZeros2[1])**(1/3.0), 0.0)
        fZeros3.append(fU + fV - fP3[2] / 3.0)
        fZeros3.append(fU * fJ + fV * fJ.conjugate() - fP3[2] / 3.0)
        fZeros3.append(fU * fJ**2 + fV * (fJ**2).conjugate() - fP3[2] / 3.0)
        return tuple(fZeros3)
    ##
    ## Solve degree 2 polynomial equation
    ## Rsolution d'une quation polynomiale de degrs 2
    ##
    def solvePoly2(self, fP2):
        """Solve polynomial equation of degree 2
        Rsolution d'une quation polynomiale de degrs 2
        @param fCoeff as tuple of float or complex with length = 2 (degree 2 coeff being always 1)
        return zeros as tuple of float or None"""
        fZeros = []
        if len(fP2) == 2:
            fP2.append(1.0)
        fDelta = fP2[1]**2 - 4 * fP2[0] * fP2[2]
        if type(fDelta) is FloatType:
            if fDelta < 0.0:
                fZeros.append(complex(-0.5 * fP2[1] / fP2[2], 0.5 * math.sqrt(abs(fDelta)) / fP2[2]))
                fZeros.append(complex(-0.5 * fP2[1] / fP2[2], -0.5 * math.sqrt(abs(fDelta)) / fP2[2]))
            else:
                fZeros.append((-fP2[1] + math.sqrt(fDelta))/(2.0 * fP2[2]))
                fZeros.append((-fP2[1] - math.sqrt(fDelta))/(2.0 * fP2[2]))
        else:
            fZeros.append((-0.5 * fP2[1] + 0.5 * cmath.sqrt(fDelta)) / fP2[2])
            fZeros.append((-0.5 * fP2[1] - 0.5 * cmath.sqrt(fDelta)) / fP2[2])
        return tuple(fZeros)
    ##
    ## Solve degree 4 polynomial equation
    ## Rsolution d'une quation polynomiale de degrs 4
    ##
    def solvePoly4(self, fCoeff):
        """Solve polynomial equation of degree 4
        Rsolution d'une quation polynomiale de degrs 4
        @param fCoeff as tuple of float with length = 5
        return zeros as tuple of float or None"""
        fZeros = None
        # ------------------------------------------------------
        if map(abs,fCoeff).count(0.0) == 5 or map(abs,fCoeff)[1:].count(0.0) == 4:
            return None
        # -------------------------------------------------
        if map(abs,fCoeff)[1:].count(0.0) == 3 and abs(fCoeff[0]) == 0.0:
            return (0.0,)
        bStop = False
        bCont = False
        # ------------------------
        if map(abs,fCoeff).count(0.0) == 3:
            for i in (1, 2, 3, 4):
                for j in range(i):
                    if abs(fCoeff[i]) != 0.0 and abs(fCoeff[j]) != 0.0:
                        bStop = True
                        fZeros = []
                        if j > 0:
                            fZeros.append(0.0)
                        fX = -fCoeff[j] / fCoeff[i]
                        if (i-j) % 2 == 0:
                            if fX > 0.0:
                                fZeros.append(fX**(1/(i-j)))
                                fZeros.append(-fX**(1/(i-j)))
                                break
                            elif not fZeros is None:
                                fZeros = None
                                break
                        else:
                            fZeros.append((1-2*(fX<0))*abs(fX)**(1/(i-j)))
                            break
                if bStop:
                    break
            else:
                return None
            if fZeros is None:
                return None
            return tuple(fZeros)
        # ------------------------
        if map(abs,fCoeff).count(0.0) == 2:
            for i in (2, 3, 4):
                for j in range(1, i):
                    for k in range(j):
                        if abs(fCoeff[i]) != 0.0 and abs(fCoeff[j]) != 0.0 and abs(fCoeff[k]) != 0.0:
                            fZeros = []
                            if k > 0:
                                bStop = True
                                fZeros.append(0.0)
                            if i-k == 2:
                                bStop = True
                                fZeros.extend(self.solvePoly2([fCoeff[k]/fCoeff[i], fCoeff[j]/fCoeff[i]]))
                            if i-k == 3:
                                bStop = True
                                fZeros.extend(self.solvePoly3([fCoeff[k]/fCoeff[i], fCoeff[j]/fCoeff[i]*((i-j)==2), fCoeff[j]/fCoeff[i]*((i-j)==1)]))
                            if i-k == 4:
                                bCont = True
                                fZeros = None
                            break
                    if bStop:
                        break
                if bStop:
                    break
            if not bCont:
                return tuple(fZeros)
        # ----------------
        if (map(abs,fCoeff)[1:].count(0.0) == 0 and abs(fCoeff[0]) == 0.0) or (map(abs,fCoeff)[:-1].count(0.0) == 0 and abs(fCoeff[-1]) == 0.0):
            fZeros = []
            if abs(fCoeff[0]) == 0.0:
                fZeros.append(0.0)
                i, j = 1, 5
            else:
                i, j = 0, 4
            fZeros.extend(self.solvePoly3(fCoeff[i:j]))
            return tuple(fZeros)
        fCoeffR = [fCoeff[i] / fCoeff[4] for i in (0,1,2,3,4)]
        # case known solution (yP null)
        fTestValue = abs(sum([(-0.5*fCoeffR[3])**i * fCoeffR[i] for i in (0,1,2,3,4)]))
        if  fTestValue == 0.0:
            if fZeros is None:
                fZeros = []
            fZeros.append(round(math.sqrt(abs(fCoeffR[1]/fCoeffR[3]))))
            fZeros.append(round(-math.sqrt(abs(fCoeffR[1]/fCoeffR[3]))))
            if 0.25 * abs(fCoeffR[3]**3/fCoeffR[1]) <= 1.0:
                fZeros.append(-0.5*fCoeffR[3])
                fZeros.append(-0.5*fCoeffR[3])
            return tuple(fZeros)
        fP43 = []
        fP43.append(-3.0 * fCoeffR[3]**4 / 256.0 + fCoeffR[3]**2 * fCoeffR[2] / 16.0 - fCoeffR[3] * fCoeffR[1] / 4.0 + fCoeffR[0]) # r
        fP43.append(fCoeffR[3]**3 / 8.0 - fCoeffR[3] / 2.0 * fCoeffR[2] + fCoeffR[1]) # q
        fP43.append(-3.0 * fCoeffR[3]**2 / 8.0 + fCoeffR[2]) # p
        # -------------------------------------
        fP3 = []
        fP3.append(0.5 * fP43[2] * fP43[0] - fP43[1]**2 / 8.0) # m
        fP3.append(-fP43[0]) # l
        fP3.append(-0.5 * fP43[2]) # k
        fRet = self.solvePoly3(fP3)
        # -------------------------------------
        fP2a = []
        fP2b = []
        fZeros = []
        for fT in fRet:
            fP2a.append(cmath.sqrt(2.0 * fT - fP43[2]))
            fP2b.append(-fP2a[-1])
            fP2a.append(fT - fP43[1] / (2.0 * fP2a[0]))
            fP2b.append(fT + fP43[1] / (2.0 * fP2a[0]))
            fP2a.reverse()
            fRes = self.solvePoly2(fP2a)
            if (fRes[0].imag == 0.0 or fRes[1].imag == 0.0):
                fTest = sum([(fRes[0].real)**i * fP43[i] for i in (0,1,2)]) + (fRes[0].real)**4
                if  abs(fTest) < 1e-6:
                    fZeros.append(fRes[0].real - fCoeffR[3] / 4.0)
                fTest = sum([(fRes[1].real)**i * fP43[i] for i in (0,1,2)]) + (fRes[1].real)**4
                if  abs(fTest) < 1e-6:
                    fZeros.append(fRes[1].real - fCoeffR[3] / 4.0)
            fP2b.reverse()
            fRes = self.solvePoly2(fP2b)
            if (fRes[0].imag == 0.0 or fRes[1].imag == 0.0):
                fTest = sum([(fRes[0].real)**i * fP43[i] for i in (0,1,2)]) + (fRes[0].real)**4
                if  abs(fTest) < 1e-6:
                    fZeros.append(fRes[0].real - fCoeffR[3] / 4.0)
                fTest = sum([(fRes[1].real)**i * fP43[i] for i in (0,1,2)]) + (fRes[1].real)**4
                if  abs(fTest) < 1e-6:
                    fZeros.append(fRes[1].real - fCoeffR[3] / 4.0)
        if fZeros is None or len(fZeros) == 0:
            return None
        return tuple(fZeros)
    ##
    ## Compute coordinates for each tangent point between an ellipse and circle
    ## Calcule les coordonnes des points de tangence entre une ellipse et un cercle
    ##
    def getTangentEllipseCircle(self, aCoord1, aSize1, fAngle1, aCoord2):
        """Compute coordinates for each tangent point between an ellipse and circle
        Calcule les coordonnes des points de tangence entre une ellipse et un cercle
        @param aCoord1 as com.sun.star.awt.Point : coordinates of ellipse center
        @param aSize1 as com.sun.star.awt.Size : half short and great axis dimension
        @param fAngle1 as float : rotation angle
        @param aCoord2 as com.sun.star.awt.Point : coordinates of circle center
        return nX1, nY1, nX2, nY2, nX3, nY3, nX4, nY4 as tuple(int, int, int, int, int, int, int, int)"""
        nTanCoord = []
        aCoordCircleCenter = {"X": 0,"Y": 0}
        aCoordCircleCenter['X'] = long(round((aCoord2["X"] - aCoord1["X"]) * math.cos(fAngle1) - (aCoord2["Y"] - aCoord1["Y"]) * math.sin(fAngle1)))
        aCoordCircleCenter['Y'] = long(round((aCoord2["X"] - aCoord1["X"]) * math.sin(fAngle1) + (aCoord2["Y"] - aCoord1["Y"]) * math.cos(fAngle1)))
        if aSize1["Width"] == aSize1["Height"]:
            if aCoordCircleCenter['X'] == 0 and aCoordCircleCenter['Y'] == 0:
                return None
            fK = aSize1["Width"] / math.sqrt(aCoordCircleCenter['X']**2 + aCoordCircleCenter['Y']**2)
            nTanCoord.append(long(round(aCoord1["X"] + fK * aCoordCircleCenter['X'])))
            nTanCoord.append(long(round(aCoord1["Y"] + fK * aCoordCircleCenter['Y'])))
            nTanCoord.append(long(round(aCoord1["X"] - fK * aCoordCircleCenter['X'])))
            nTanCoord.append(long(round(aCoord1["Y"] - fK * aCoordCircleCenter['Y'])))
            return tuple(nTanCoord)
        fC112 = float(aSize1["Width"]**2 - aSize1["Height"]**2)
        # ---------------------------
        if aCoordCircleCenter['X'] == 0:
            nTanCoord.append(long(round(aCoord1["X"] + aSize1["Height"] * math.sin(fAngle1))))
            nTanCoord.append(long(round(aCoord1["Y"] + aSize1["Height"] * math.cos(fAngle1))))
            nTanCoord.append(long(round(aCoord1["X"] - aSize1["Height"] * math.sin(fAngle1))))
            nTanCoord.append(long(round(aCoord1["Y"] - aSize1["Height"] * math.cos(fAngle1))))
            fY = -float(aSize1["Height"]**2*aCoordCircleCenter['Y'])/fC112
            if (fY/aSize1["Height"])**2 <= 1.0:
                fX = aSize1["Width"]*math.sqrt(1.0-(fY/aSize1["Height"])**2)
                nTanCoord.append(long(round(aCoord1["X"] + fX * math.cos(fAngle1) + fY * math.sin(fAngle1))))
                nTanCoord.append(long(round(aCoord1["Y"] - fX * math.sin(fAngle1) + fY * math.cos(fAngle1))))
                nTanCoord.append(long(round(aCoord1["X"] - fX * math.cos(fAngle1) + fY * math.sin(fAngle1))))
                nTanCoord.append(long(round(aCoord1["Y"] + fX * math.sin(fAngle1) + fY * math.cos(fAngle1))))
            return tuple(nTanCoord)
        # ---------------------------
        if aCoordCircleCenter['Y'] == 0:
            nTanCoord.extend([long(round(aCoord1["X"] + aSize1["Width"] * math.cos(fAngle1))), long(round(aCoord1["Y"] - aSize1["Width"] * math.sin(fAngle1)))])
            nTanCoord.extend([long(round(aCoord1["X"] - aSize1["Width"] * math.cos(fAngle1))), long(round(aCoord1["Y"] + aSize1["Width"] * math.sin(fAngle1)))])
            if (aSize1["Width"] * aCoordCircleCenter['X'] / fC112)**2 < 1.0:
                fX = aSize1["Width"]**2 * aCoordCircleCenter['X'] / fC112
                fY = aSize1["Height"] * math.sqrt(1 - (aSize1["Width"] * aCoordCircleCenter['X'] / fC112)**2)
                nTanCoord.append(long(round(aCoord1["X"] + fX * math.cos(fAngle1) + fY * math.sin(fAngle1))))
                nTanCoord.append(long(round(aCoord1["Y"] - fX * math.sin(fAngle1) + fY * math.cos(fAngle1))))
                nTanCoord.append(long(round(aCoord1["X"] + fX * math.cos(fAngle1) - fY * math.sin(fAngle1))))
                nTanCoord.append(long(round(aCoord1["Y"] - fX * math.sin(fAngle1) - fY * math.cos(fAngle1))))
            return tuple(nTanCoord)
        # -----------------------------------
        try:
            fK = math.pow(10, -max(round(abs(math.log10(aSize1["Width"]))), round(abs(math.log10(aSize1["Height"])))))
        except:
            fk = 1.0
        fCoeff = []
        fCoeff.append(-float(aCoordCircleCenter['X'] * fK * (aSize1["Width"] * fK)**3)**2)
        fCoeff.append(2.0 * (aSize1["Width"] * fK)**4 * aCoordCircleCenter['X'] * fK**3 * fC112)
        fCoeff.append((aSize1["Width"] * fK)**2 * ((aSize1["Width"] * aCoordCircleCenter['X'] * fK**2)**2 + (aSize1["Height"] * aCoordCircleCenter['Y'] * fK**2)**2 - (fC112 * fK**2)**2))
        fCoeff.append(-2.0 * (aSize1["Width"] * fK)**2 * aCoordCircleCenter['X'] * fC112 * fK**3)
        fCoeff.append((fC112 * fK**2)**2)
        # -------------------------------------
        fTanCoordX = self.solvePoly4(fCoeff)
        if fTanCoordX is None:
            return None
        for fXr in fTanCoordX:
            fX = fXr / fK
            fY = aSize1["Height"]**2 * aCoordCircleCenter['Y'] * fX / (aSize1["Width"]**2 * aCoordCircleCenter['X'] - fC112 * fX)
            nTanCoord.append(long(round(aCoord1["X"] + fX * math.cos(fAngle1) + fY * math.sin(fAngle1))))
            nTanCoord.append(long(round(aCoord1["Y"] - fX * math.sin(fAngle1) + fY * math.cos(fAngle1))))
        return tuple(nTanCoord)
    ##
    ## Compute coordinates for each tangent point between an ellipse and a line
    ## Calcule les coordonnes des points de tangence entre une ellipse et une droite
    ##
    def getTangentEllipseLine(self, aCoord1, aSize1, fAngle1, aCoord2):
        """Compute coordinates for each tangent point between an ellipse and line
        Calcule les coordonnes des points de tangence entre une ellipse et une droite
        @param aCoord1 as com.sun.star.awt.Point : coordinates of ellipse center
        @param aSize1 as com.sun.star.awt.Size : half short and great axis dimension
        @param fAngle1 as float : rotation angle
        @param aCoord2 as com.sun.star.awt.Point : coordinates of a point belonging to the line
        return nX1, nY1, nX2, nY2 as tuple(int, int, int, int)"""
        nTanCoord = []
        aCoord = {'X': 0, 'Y': 0}
        aCoordLine = {"X": 0, "Y": 0}
        aCoordLine['X'] = long(round((aCoord2["X"] - aCoord1["X"]) * math.cos(fAngle1) - (aCoord2["Y"] - aCoord1["Y"]) * math.sin(fAngle1)))
        aCoordLine['Y'] = long(round((aCoord2["X"] - aCoord1["X"]) * math.sin(fAngle1) + (aCoord2["Y"] - aCoord1["Y"]) * math.cos(fAngle1)))
        nDelta = aSize1["Width"]**2 * aCoordLine['Y']**2 + aSize1["Height"]**2 * aCoordLine['X']**2 - aSize1["Width"]**2 * aSize1["Height"]**2
        if nDelta < 0:
            return None
        elif nDelta == 0:
            return (aCoord2["X"], aCoord2["Y"])
        fM1 = (-aCoordLine['X'] * aCoordLine['Y'] + math.sqrt(nDelta)) / (aSize1["Width"]**2 - aCoordLine['X']**2)
        fP1 = aCoordLine['Y'] - fM1 * aCoordLine['X']
        fM2 = (-aCoordLine['X'] * aCoordLine['Y'] - math.sqrt(nDelta)) / (aSize1["Width"]**2 - aCoordLine['X']**2)
        fP2 = aCoordLine['Y'] - fM2 * aCoordLine['X']
        aCoord["X"] = long(round(-aSize1["Width"]**2 * fM1 * fP1 / (aSize1["Height"]**2 + aSize1["Width"]**2 * fM1**2)))
        aCoord["Y"] = long(round(fM1 * aCoord["X"] + fP1))
        nTanCoord.append(long(round(aCoord1["X"] + aCoord["X"] * math.cos(fAngle1) + aCoord["Y"] * math.sin(fAngle1))))
        nTanCoord.append(long(round(aCoord1["Y"] - aCoord["X"] * math.sin(fAngle1) + aCoord["Y"] * math.cos(fAngle1))))
        aCoord["X"] = long(round(-aSize1["Width"]**2 * fM2 * fP2 / (aSize1["Height"]**2 + aSize1["Width"]**2 * fM2**2)))
        aCoord["Y"] = long(round(fM2 * aCoord["X"] + fP2))
        nTanCoord.append(long(round(aCoord1["X"] + aCoord["X"] * math.cos(fAngle1) + aCoord["Y"] * math.sin(fAngle1))))
        nTanCoord.append(long(round(aCoord1["Y"] - aCoord["X"] * math.sin(fAngle1) + aCoord["Y"] * math.cos(fAngle1))))
        return tuple(nTanCoord)
    ##
    ## Compute coordinates of circle center and its radius given 3 points
    ## Calcule les coordonnees du centre et son rayon d'un cercle passant par 3 points
    ##
    def getCircle3P(self, aCoordA, aCoordB, aCoordM):
        """Compute coordinates of circle center and its radius given 3 points
        Calcule les coordonnees du centre et son rayon d'un cercle passant par 3 points
        @param aCoord1 as com.sun.star.awt.Point : coordinates of a point belonging to the circle
        @param aCoord2 as com.sun.star.awt.Point : coordinates of a point belonging to the circle
        @param aCoord3 as com.sun.star.awt.Point : coordinates of a point belonging to the circle
        return nXCenter, nYCenter, nRadius as tuple(int, int, int)"""
        nCoeff = [[aCoordB["X"] - aCoordA["X"], aCoordB["Y"] - aCoordA["Y"], (aCoordB["X"]**2 - aCoordA["X"]**2 + aCoordB["Y"]**2 - aCoordA["Y"]**2)/2],]
        nCoeff.append([aCoordM["X"] - aCoordA["X"], aCoordM["Y"] - aCoordA["Y"], (aCoordM["X"]**2 - aCoordA["X"]**2 + aCoordM["Y"]**2 - aCoordA["Y"]**2)/2])
        nDet = nCoeff[0][0] * nCoeff[1][1] - nCoeff[1][0] * nCoeff[0][1]
        nCircleData = []
        if nDet == 0:
            nCircleData.append((aCoordA["X"] + aCoordB["X"]) / 2)
            nCircleData.append((aCoordA["Y"] + aCoordB["Y"]) / 2)
        else:
            nCircleData.append((nCoeff[0][2] * nCoeff[1][1] - nCoeff[1][2] * nCoeff[0][1]) / nDet)
            nCircleData.append((nCoeff[0][0] * nCoeff[1][2] - nCoeff[1][0] * nCoeff[0][2]) / nDet)
        nCircleData.append(long(round(math.sqrt((nCircleData[0] - aCoordA["X"])**2 + (nCircleData[1] - aCoordA["Y"])**2))))
        return tuple(nCircleData)
    ##
    ## Compute X, Y from absissa, ordinate of a point related to a line
    ## Calcule X,Y  partir de absisse, ordonnee par rapport  un droite d'un point
    ##
    def getCoordFromAbsOrd(self, aCoordA, aCoordB, aAbsissa, aOrdinate):
        """Compute X, Y from absissa, ordinate of a point related to a line
        Calcule X,Y �partir de absisse, ordonnee par rapport �un droite d'un point
        @param aCoordA as com.sun.star.awt.Point : origin coordinates of the line
        @param aCoordB as com.sun.star.awt.Point : end coordinates of the line
        @param aAbsissa as Long
        @param aOrdinate as long
        return coordinates as com.sun.star.awt.Point"""
        nDist, fAngle = self.toPolarCoord(aCoordA, aCoordB)
        aCoord = {"X": 0,"Y": 0}
        aCoord["X"] = aCoordA["X"] + long(round(aAbsissa * math.cos(fAngle) + aOrdinate * math.sin(fAngle)))
        aCoord["Y"] = aCoordA["Y"] + long(round(-aAbsissa * math.sin(fAngle) + aOrdinate * math.cos(fAngle)))
        return aCoord
    ##
    ## Compute absissa, ordinate from a point to a line and the length of the line
    ## Calcule absisse et ordonnee d'un point a un segment et la longueur
    ##
    def getDist2Line(self, aCoordA, aCoordB, aCoordM):
        """Compute a distance from a point to a line
        Calcule la distance d'un point a une droite
        @param aCoordA as com.sun.star.awt.Point : coordinates of a point belonging to the line
        @param aCoordB as com.sun.star.awt.Point : coordinates of a point belonging to the line
        @param aCoordC as com.sun.star.awt.Point : coordinates of a point off line
        return absissa, ordinate, length, projection coordinates as tuple(long, long, long, com.sun.star.awt.Point)"""
        aPoint = {"X": 0,"Y": 0}
        nArea = (aCoordB["Y"] - aCoordM["Y"]) * aCoordA["X"] + (aCoordM["Y"] - aCoordA["Y"]) * aCoordB["X"] + (aCoordA["Y"] - aCoordB["Y"]) * aCoordM["X"]
        fDist = math.hypot(float(aCoordA["X"] - aCoordB["X"]), (aCoordA["Y"] - aCoordB["Y"]))
        if fDist > 1.0:
            nProd = (aCoordB["X"] - aCoordA["X"]) * (aCoordM["X"] - aCoordA["X"]) + (aCoordB["Y"] - aCoordA["Y"]) * (aCoordM["Y"] - aCoordA["Y"])
            aPoint["X"] = aCoordA["X"] + long(round(nProd * (aCoordB["X"] - aCoordA["X"]) / fDist**2))
            aPoint["Y"] = aCoordA["Y"] + long(round(nProd * (aCoordB["Y"] - aCoordA["Y"]) / fDist**2))
            return long(round(nProd / fDist)), long(round(nArea / fDist)), long(round(fDist)), aPoint
        else:
            nDist = self.toPolarCoord(aCoordA, aCoordM)[0]
            return 0, nDist, 0, aCoordA
    ##
    ## Compute absissa, odinate from a point to an arc and the length of arc
    ## Calcule absisse et ordonnee d'un point a un arc et la longueur de l arc
    ##
    def getDist2Arc(self, aCoordA, aSize, fRotation, fStartAngle, fEndAngle, aCoordB):
        """Compute absissa, ordinate from a point to an arc and the length of arc
        Calcule absisse et ordonnee d'un point a un arc et la longueur de l arc
        @param aCoordA as com.sun.star.awt.Point : coordinates of arc center
        @param aSize as com.sun.star.awt.Size : radius of arc
        @param fRotation as float : rotate angle
        @param fStartAngle as float : start angle of arc
        @param fEndAngle as float : end angle of arc
        @param aCoordB as com.sun.star.awt.Point : coordinates of a point off arc
        return absissa, ordinate, length, projection coordinates  as tuple(long, long, long, com.sun.star.awt.Point)"""
        nDist2Arc = [0, 0, 0]
        aPoint = {"X": 0,"Y": 0}
        # ellipse is quite a circle
        if abs(aSize["Width"] - aSize["Height"]) < 5:
            fEndAlfa = fEndAngle - fStartAngle
            if fEndAlfa < 0:
                fEndAlfa = 2 * math.pi + fEndAlfa
            fPointAlfa = -math.atan2(aCoordB["Y"]-aCoordA["Y"], aCoordB["X"]-aCoordA["X"]) - fStartAngle
            if fPointAlfa < 0:
                fPointAlfa = fPointAlfa + 2 * math.pi
                if fPointAlfa < 0:
                    fPointAlfa = fPointAlfa + 2 * math.pi
            if fPointAlfa >= math.pi + fEndAlfa / 2:
                bNegAbsisse = 1
            else:
                bNegAbsisse = 0
            nDist2Arc[0] = long(round(aSize["Width"]*(fPointAlfa-bNegAbsisse*2*math.pi)))
            nDist2Arc[1] = long(round(math.hypot(aCoordB["X"]-aCoordA["X"], aCoordB["Y"]-aCoordA["Y"])-aSize["Width"]))
            nDist2Arc[2] = long(round(fEndAlfa*aSize["Width"]))
            aPoint["X"] = aCoordA["X"] + aSize["Width"] * math.cos(fStartAngle + fPointAlfa)
            aPoint["Y"] = aCoordA["Y"] - aSize["Width"] * math.sin(fStartAngle + fPointAlfa)
            nDist2Arc.append(aPoint)
            return tuple(nDist2Arc)
        # transform global coordinates to arc local ones
        aLocalPoint = {"X": 0,"Y": 0}
        aLocalPoint["X"] = long(round((aCoordB["X"] - aCoordA["X"]) * math.cos(fRotation) + (aCoordB["Y"] - aCoordA["Y"]) * math.sin(-fRotation)))
        aLocalPoint["Y"] = long(round(-(aCoordB["X"] - aCoordA["X"]) * math.sin(-fRotation) + (aCoordB["Y"] - aCoordA["Y"]) * math.cos(fRotation)))
        # compute ordinate sign
        fSignOrd =  (float(aLocalPoint["X"]) / aSize["Width"])**2 + (float(aLocalPoint["Y"]) / aSize["Height"])**2
        if fSignOrd - 1 > 1e-1:
            bSignOrd = 1
        elif fSignOrd - 1 < -1e-1:
            bSignOrd = -1
        else:
            bSignOrd = 0
        # general case
        if bSignOrd != 0:
            ret = self.getTangentEllipseCircle(aCoordA, aSize, fRotation, aCoordB)
            if ret is None:
                fCoords = [aCoordA["X"] + aSize["Width"], aCoordA["Y"]]
            else:
                fCoords = list(ret)
                nDist2Arc[1] = long(round(math.hypot(aCoordB["X"]-aCoordA["X"], aCoordB["Y"]-aCoordA["Y"])))
                nI = 0
                for i in range(len(fCoords)/2):
                    nD = long(round(math.hypot(fCoords[2*i] - aCoordB["X"], fCoords[2*i+1] - aCoordB["Y"])))
                    if nD < nDist2Arc[1]:
                        nI = i
                        nDist2Arc[1] = nD
                aPoint["X"], aPoint["Y"] = long(round(fCoords[2*nI])), long(round(fCoords[2*nI+1]))
        else:
            aPoint["X"], aPoint["Y"] = aCoordB["X"], aCoordB["Y"]
        nDist2Arc[1] = bSignOrd * nDist2Arc[1]
        # transform parametric angles to polar ones
        fStartTheta = math.atan2(aSize["Height"] * math.sin(fStartAngle), aSize["Width"] * math.cos(fStartAngle))
        if fStartTheta < 0:
            fStartTheta = fStartTheta + 2 * math.pi
        fEndAlfa = math.atan2(aSize["Height"] * math.sin(fEndAngle), aSize["Width"] * math.cos(fEndAngle))
        if fEndAlfa < 0:
            fEndAlfa = fEndAlfa + 2 * math.pi
        fEndAlfa = fEndAlfa - fStartTheta
        # Start angle must be less than end one
        if fEndAlfa < 0:
            fEndAlfa = 2 * math.pi + fEndAlfa
        nDist2Arc[2] = long(round(fEndAlfa * math.sqrt(aSize["Width"]*aSize["Height"])))
        # get local polar angle of extern point
        fPointAlfa = -math.atan2(aPoint["Y"]-aCoordA["Y"], aPoint["X"]-aCoordA["X"]) - fRotation - fStartTheta
        while fPointAlfa < 0:
            fPointAlfa = fPointAlfa + 2 * math.pi
        if fPointAlfa >= math.pi + fEndAlfa / 2:
            nDist2Arc[0] = long(round((fPointAlfa - 2 * math.pi) * math.sqrt(aSize["Width"]*aSize["Height"])))
        else:
            nDist2Arc[0] = long(round(fPointAlfa * math.sqrt(aSize["Width"]*aSize["Height"])))
        nDist2Arc.append(aPoint)
        return tuple(nDist2Arc)
    ##
    ## Compute the area of an ellipse pie
    ##
    def getArcEllipseArea(self, nWidth, nHeight, fStartAngle, fEndAngle, sType):
        """
        Compute the area of an ellipse pie
        Calcule l'aire d'un secteur d'ellipse
        @param nWidth as integer : ellipse length axis 1
        @param nHeight as integer : ellipse length axis 2
        @param fStartAngle as float : start angle of the arc
        @param fEndAngle as float : end angle of the arc
        @sType as string : run direction
        @return fArea as float: area of ellipse pie
        """
        if sType == "A":
            fTheta0 = fStartAngle
            fTheta1 = fEndAngle
            fInc = -0.01
            if fStartAngle > fEndAngle:
                fTheta1 += (2 * math.pi)
        else:
            fTheta0 = fEndAngle
            fTheta1 = fStartAngle
            fInc = 0.01
            if fStartAngle < fEndAngle:
                fTheta1 += (2 * math.pi)
        fAlfas = [fTheta0]
        fAlfas.append(min(fTheta1, (int(2 * fTheta0 / math.pi) + 1) * math.pi / 2))
        if fAlfas[-1] < fTheta1:
            fAlfas.append(int(2 * fTheta1 / math.pi) * math.pi / 2)
        else:
            fAlfas.append(fTheta1)
        fAlfas.append(fTheta1)
        #self.showDebugMsg("angles : %.4f %.4f %.4f %.4f" % tuple(fAlfas))
        fArea = 0.0
        while fTheta0 < fAlfas[1] - abs(fInc):
            fRho1 = nWidth * nHeight / math.sqrt((nHeight * math.cos(fTheta0))**2 + (nWidth * math.sin(fTheta0))**2)
            fRho2 = nWidth * nHeight / math.sqrt((nHeight * math.cos(fTheta0 + abs(fInc)))**2 + (nWidth * math.sin(fTheta0 + abs(fInc)))**2)
            fArea = fArea + fInc * (fRho1**2 + fRho2**2) / 4
            fTheta0 += abs(fInc)
        #self.showDebugMsg("Area = %.4f" % fArea)
        fArea = fArea + nWidth * nHeight * (fAlfas[2] - fAlfas[1]) / 2
        #self.showDebugMsg("Area = %.4f" % fArea)
        fTheta0 = fAlfas[2]
        while fTheta0 < fAlfas[3] - abs(fInc):
            fRho1 = nWidth * nHeight / math.sqrt((nHeight * math.cos(fTheta0))**2 + (nWidth * math.sin(fTheta0))**2)
            fRho2 = nWidth * nHeight / math.sqrt((nHeight * math.cos(fTheta0 + abs(fInc)))**2 + (nWidth * math.sin(fTheta0 + abs(fInc)))**2)
            fArea = fArea + fInc * (fRho1**2 + fRho2**2) / 4
            fTheta0 += abs(fInc)
        #self.showDebugMsg("Area = %.4f" % fArea)
        return fArea
    ##
    ## Compute absissa, ordinate of a point of an ellipse or circle
    ## Calcule absisse et ordonnee d'un point d'une ellipse ou cercle
    ##
    def getPointOnEllipse(self, fPolarAngle, fEllAngle, aCenter, aSize, fRotation):
        """Compute absissa, ordinate of a point of an ellipse or circle
        Calcule absisse et ordonnee d'un point d'une ellipse ou cercle
        @param fPolarAngle as float : polar global angle
        @param fEllAngle as float : elliptic local angle
        @param aCenter as com.sun.star.awt.Point : center coordinates
        @param aSize as com.sun.star.awt.Size : axis sizes of ellipse
        @param fRotation as float : rotation angle
        @return tuple (aGlobalPoint as com.sun.star.awt.Point, fTheta as float)"""
        if aSize["Width"] == 0:
            return aCenter, float(0 + fRotation)
        if aSize["Height"] == 0:
            return aCenter, math.pi/2 + fRotation
        aLocalPoint = [{"X": 0,"Y": 0}, {"X": 0,"Y": 0}]
        aGlobalPoint = {"X": 0,"Y": 0}
        aLocalPoint[1]["X"] = aLocalPoint[1]["Y"] = 0L
        if not fPolarAngle is None:
            fAlfa = fPolarAngle - fRotation
            aLocalPoint[0]["X"] = long(round(aSize["Width"] * aSize["Height"] * math.cos(fAlfa) / math.sqrt((aSize["Width"] * math.sin(-fAlfa))**2+(aSize["Height"] * math.cos(fAlfa))**2)))
            aLocalPoint[0]["Y"] = long(round(aSize["Width"] * aSize["Height"] * math.sin(-fAlfa) / math.sqrt((aSize["Width"] * math.sin(-fAlfa))**2+(aSize["Height"] * math.cos(fAlfa))**2)))
            fTheta = math.atan2(aSize["Width"] * math.sin(fAlfa), aSize["Height"] * math.cos(fAlfa))
            if fTheta < 0:
                fTheta = fTheta + math.pi * 2
        else:
            aLocalPoint[0]["X"] = long(round(aSize["Width"] * math.cos(fEllAngle)))
            aLocalPoint[0]["Y"] = long(round(aSize["Height"] * math.sin(-fEllAngle)))
            fTheta = math.fmod(self.toPolarCoord(aLocalPoint[1], aLocalPoint[0])[1] + fRotation, 2 * math.pi)
        aGlobalPoint["X"] = long(round(aCenter["X"] + aLocalPoint[0]["X"] * math.cos(fRotation) + aLocalPoint[0]["Y"] * math.sin(fRotation)))
        aGlobalPoint["Y"] = long(round(aCenter["Y"] - aLocalPoint[0]["X"] * math.sin(fRotation) + aLocalPoint[0]["Y"] * math.cos(fRotation)))
        return aGlobalPoint, fTheta
    ##
    ## Compute the angle of a tangent at a given point on an ellipse or circle
    ## Calcule l'angle de la tangente en un point donn d'une ellipse ou cercle
    ##
    def getTangentAngle(self, aCenter, aSize, fRotation, aPoint):
        """Compute the angle of a tangent at a given point on an ellipse or circle
        Calcule l'angle de la tangente en un point donn d'une ellipse ou cercle
        @param aCenter as com.sun.star.awt.Point : center coordinates
        @param aSize as com.sun.star.awt.Size : axis sizes of ellipse
        @param fRotation as float : rotation angle
        @param aPoint as com.sun.star.Point : a point on the ellipse
        @return fTheta as float"""
        if aSize["Width"] == 0 or aSize["Height"] == 0:
            return None
        aLocalPoint = {"X": 0,"Y": 0}
        aLocalPoint["X"] = long(round((aPoint["X"] - aCenter["X"]) * math.cos(fRotation) - (aPoint["Y"] - aCenter["Y"]) * math.sin(fRotation)))
        aLocalPoint["Y"] = long(round((aPoint["X"] - aCenter["X"]) * math.sin(fRotation) + (aPoint["Y"] - aCenter["Y"]) * math.cos(fRotation)))
        fTest = (float(aLocalPoint["X"])/aSize["Width"])**2 + (float(aLocalPoint["Y"])/aSize["Height"])**2 -1.0
        if  abs(fTest) > 1e-2:
            print fTest
            return None
        fTheta = math.fmod(math.atan2(aSize["Height"]**2 * aLocalPoint["X"], aSize["Width"]**2 * aLocalPoint["Y"]) + fRotation, 2*math.pi)
        if fTheta < 0:
            fTheta = fTheta + 2 * math.pi
        return fTheta
