# $Id$
# -*- coding: latin-1 -*-
## ********************************************************************************
## entitiespot vendredi 6 mars 2009, 14:58:09 (UTC+0100)

## Copyright (C) 2005-2009 Gerard Deneux
## gerard.deneux@free.fr

## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.

## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
## Lesser General Public License for more details.
## http://www.opensource.org/licenses/lgpl-license.php

## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
## ******************************************************************************
try:
    import math
    from types import *
    from com.sun.star.drawing.CircleKind import ARC, SECTION, CUT, FULL
except:
    print "entitiespot : probleme d'import"

# ===============================================
# Indique les entites graphique sous le pointeur
# shows the graphical entities under pointer
# ===============================================
class EntitieSpot():
    """Indique les entites graphique sous le pointeur
    shows the graphical entities under pointer"""
    def __init__(self, ctx):
        import fr.free.deneux.gerard.cadooo.compute
        self.ctx = ctx
        self.xCompute = fr.free.deneux.gerard.cadooo.compute.Compute(ctx)
        self.aEntityList = []
        self.bValidList = False
        self.aSpotList = None
        self.aSpottedShape = None
        self.aMouseRelPos = None
        self.aBorderList = None
        self.aSplittedShape = None
        self.aEntityEnvAndCoord = {}
    ##
    ## object inspection wrapper
    ##
    def inspectObject(self, aObject):
        if not hasattr(self, "xDiveIn"):
            from fr.free.deneux.gerard.cadooo.divein import DiveIn
            self.xDiveIn = DiveIn(self.ctx)
        ret = self.xDiveIn.execute(aObject)
    # ========================================
    # Look for entities under mouse
    # ========================================
    def execute(self, aArgs):
        self.aEnv = {}
        self.aEnv.update(aArgs)
        # --------------------------------------------------------------------
        if not self.bValidList: # create a list of splitted entities
            # EntityList definition
            # Line = entity address (indexes tuple), entity type, Point, Point
            # ellipse = entity address (indexes tuple), entity type, Point, Size, RotateAngle, StartAngle, EndAngle
            if self.createEntityList(self.aEnv['CONTROLLER']):
                self.bValidList = True
            else:
                self.aSpotList = None
                self.aSpottedShape = None
                self.aBorderList = None
                self.aSplittedShape = None
                self.aEntityEnvAndCoord = {'UNOTYPE':None, 'ENTITYPE':None, 'STEP':0, 'MODE':[self.aEnv['MODE'][0]], 'COORDS':None}
                return self.aEnv
        if self.getSpottedEntities():
            self.setEntityEnvAndCoord(spottedShape=self.aSpottedShape)
            if not self.aEnv['EDITMODE'] is None:
                self.getBorders(self.aMouseRelPos)
        return self.aEnv
    #
    # Create a list of visible or selected entity that are splitted
    #
    def createEntityList(self, xController, bFromSelection=False):
        nIndexes = []
        aShapes = []
        aShapes.append(xController.getSelection())
        if aShapes[-1] is None:
            if not bFromSelection:
                aShapes[-1] = xController.getCurrentPage()
            else:
                return False
        nEntityCount = [aShapes[-1].getCount()]
        aVisibleArea = xController.VisibleArea
        aCoord = ({"X": 0,"Y": 0}, {"X": 0,"Y": 0})
        aCoord[0]["X"], aCoord[0]["Y"] = aVisibleArea.X + aVisibleArea.Width / 2, aVisibleArea.Y + aVisibleArea.Height / 2
        # ----------------------
        if nEntityCount[-1] > 0:
            nIndexes.append(0)
            # --------
            while True:
                nShapeID = []
                aShape = aShapes[-1].getByIndex(nIndexes[-1])
                nShapeID.append(aShape.ZOrder)
                aParent = aShape.getParent()
                while True:
                    if hasattr(aParent, "supportsService"):
                        if aParent.supportsService("com.sun.star.drawing.GroupShape"):
                            nShapeID.append(aParent.ZOrder)
                            aParent = aParent.getParent()
                        else:
                            break
                nShapeID.reverse()
                aType = aShape.getShapeType().split('.')[-1]
                # -----------------------------------------------------------------------------------------------
                if aType in ("PolyLineShape", "PolyPolygonShape", "RectangleShape", "LineShape", "EllipseShape"):
                    # ---------------------------------------
                    if not aShape.getName() in ("new", "ucs"): # do not treat the shape being created
                        aCoord[1]["X"] = aShape.BoundRect.X + aShape.BoundRect.Width / 2
                        aCoord[1]["Y"] = aShape.BoundRect.Y + aShape.BoundRect.Height / 2
                        nWidth = aVisibleArea.Width + aShape.BoundRect.Width
                        nHeight = aVisibleArea.Height + aShape.BoundRect.Height
                        nRateX = 200 * abs(aCoord[1]["X"] - aCoord[0]["X"]) / nWidth
                        nRateY = 200 * abs(aCoord[1]["Y"] - aCoord[0]["Y"]) / nHeight
                        nMaxRateX = 100 * aVisibleArea.Width / nWidth
                        nMaxRateY = 100 * aVisibleArea.Height / nHeight
                        # PolyPolygonBezierShape and other shapes are excluded (todo)
                        if nRateX <= nMaxRateX and nRateY <= nMaxRateY:
                            self.aEntityList.extend(self.splitShape(aShape, aType, nShapeID))
                # -----------------------------------------------------------------------
                elif aType in ("GroupShape",) and not aShape.getName() in ("dimension",): # add a new level to treat grouped shapes
                    aShapes.append(aShape)
                    nEntityCount.append(aShape.getCount())
                    nIndexes.append(-1)
                # ----------------------
                while len(nIndexes) > 0:
                    nIndexes[-1] = nIndexes[-1] + 1
                    # ----------------------------------
                    if nIndexes[-1] >= nEntityCount[-1]: # level of grouped shapes completed
                        nIndexes.pop()
                        aShapes.pop()
                        nEntityCount.pop()
                    else:
                        break # still some shapes to split
                else:
                    break # all shapes done
        if len(self.aEntityList) > 0:
            return True
        return False
    #
    # get spotted entities
    #
    def getSpottedEntities(self, aArgs=None):
        if not aArgs is None:
            if not hasattr(self, "aEnv"):
                self.aEnv = {}
            self.aEnv.update(aArgs)
        nFieldLength = self.aEnv['MAGNETSIZE'] * self.aEnv['DEVICEXSCALE']
        aMouseParm = dict([(k, self.aEnv[k]) for k in self.aEnv['MOUSEPARM']])
        aMouseCoord = self.xCompute.transformMouseCoord(aMouseParm)
        self.aSpottedShape = None
        self.aSplittedShape = None
        aSegments = []
        aDistList =[] # absissa, ordinate, length, projection point
        aIndxList = []
        nSubIndx = 1
        nI = nDist = -1
        bEntityAdded = False
        aCoord = ({"X": 0,"Y": 0}, {"X": 0,"Y": 0})
        # ------------------------------------
        for aEntity in self.aEntityList:
            # ------------------------------------------
            if aEntity[1] in ("LineShape",): # line spotted
                aDistList.append(self.xCompute.getDist2Line(aEntity[2], aEntity[3], aMouseCoord))
                # ------------------------------------------
                if abs(aDistList[-1][1]) <= nFieldLength:
                    if aDistList[-1][0]+ nFieldLength >= 0 and aDistList[-1][0] - nFieldLength <= aDistList[-1][2]:
                        aSegments.append(aEntity)
                        bEntityAdded = True
            # ---------------------------------------------- # ellipse spotted
            elif aEntity[1] in ("EllipseShape", "ArcEllipseShape", "ArcCircleShape", "CircleShape"):
                aDistList.append(self.xCompute.getDist2Arc(aEntity[2], aEntity[3], aEntity[4], aEntity[5], aEntity[6], aMouseCoord))
                # -----------------------------------------
                if abs(aDistList[-1][1]) <= nFieldLength:
                    if aDistList[-1][0] + nFieldLength >= 0 and aDistList[-1][0] - nFieldLength <= aDistList[-1][2]:
                        aSegments.append(aEntity)
                        bEntityAdded = True
            # ------------------------------------------
            elif aEntity[1].count("Centre") > 0:
                nRho = long(round(math.hypot(aEntity[2]['X'] - aMouseCoord["X"], aEntity[2]['Y'] - aMouseCoord["Y"])))
                aDistList.append((0, nRho, 0, {"X": 0,"Y": 0}))
                if nRho <= nFieldLength:
                    aSegments.append(aEntity)
            # --------------------------------------------
            if aEntity[1].count("Centre") == 0:
                if aIndxList == aEntity[0]:
                    nSubIndx = nSubIndx + 1
                else:
                    aIndxList = aEntity[0]
                    nSubIndx = 1
            # --------------------------------------------------------
            if bEntityAdded:
                bEntityAdded = False
                if self.aSpottedShape is None or abs(aDistList[-1][1]) < nDist:
                    self.aSpottedShape = []
                    for aItem in aEntity:
                        self.aSpottedShape.append(aItem)
                    self.aSpottedShape.append(nSubIndx)
                    self.aMouseRelPos = aDistList[-1]
                    nDist = abs(aDistList[-1][1])
        self.aEnv['GROUPSHAPE'] = self.aEnv['CONTROLLER'].getCurrentPage()
        if len(aSegments) > 0:
            self.aSpotList = []
            for aItem in aSegments:
                self.aSpotList.append([])
                self.aSpotList[-1].extend(aItem)
            if not self.aSpottedShape is None:
                if len(self.aSpottedShape[0]) > 1:
                    for j in self.aSpottedShape[0][:-1]:
                        self.aEnv['GROUPSHAPE'] = self.aEnv['GROUPSHAPE'].getByIndex(j)
                j = self.aSpottedShape[0][-1]
                aShape = self.aEnv['GROUPSHAPE'].getByIndex(j)
                aType = aShape.getShapeType().split('.')[-1]
                self.aSplittedShape = self.splitShape(aShape, aType, self.aSpottedShape[0])
            else:
                return False
            return True
        else:
            self.aMouseRelPos = None
            self.aSpotList = None
            self.aSpottedShape = None
            self.aSplittedShape = None
            return False
    #
    # get splitted entity[ies]
    #
    def getSplittedEntity(self, xShape=None, aSpottedShape=None):
        aSplittedShape = []
        if xShape is None:
            for aItem in self.aEntityList:
                aSplittedShape.append(aItem)
        else:
            sType = xShape.getShapeType().split('.')[-1]
            # -----------------------------------------------------------------------------------------------
            if sType in ("PolyLineShape", "PolyPolygonShape", "RectangleShape", "LineShape", "EllipseShape"):
                aSplittedShape = self.splitShape(xShape, sType, aSpottedShape[0])
        if len(aSplittedShape) > 0:
            return aSplittedShape
        return None
    #
    # get borders
    #
    def getBorders(self, aMouseRelPos):
        self.aBorderList = {}
        aCoord = ({"X": 0,"Y": 0}, {"X": 0,"Y": 0})
        for aKey in ('IOB', 'OOB', 'IEB', 'OEB'):
            self.aBorderList[aKey] = {'DELTA':-1, 'RATE':-1, "X": 0,"Y": 0}
        aSpottedId = self.aSpottedShape[0] + (self.aSpottedShape[-1],)
        bReverse = (self.aSpottedShape[1] in ("CircleShape", "EllipseShape"), self.aSpottedShape[1] in ("ArcCircleShape", "ArcEllipseShape"))
        if bReverse[0] or bReverse[1]:
            self.aReverseBorderList = {}
            for aKey in ('IOB', 'OOB', 'IEB', 'OEB'):
                self.aReverseBorderList[aKey] = {'DELTA':-1, 'RATE':-1, "X": 0,"Y": 0}
        # -----------------------------
        for aEntity in self.aEntityList:
            # --------------------------------------------------------
            if aSpottedId != aEntity[0] + (aEntity[-1],): # entity cannot be self border
                # ----------------------------------------
                if self.aSpottedShape[1] == "LineShape": # spotted entity is a line
                    # ---------------------------------------
                    if aEntity[1] == "LineShape": # border candidate is a line
                        aResult = self.xCompute.getInterLines2(self.aSpottedShape[2], self.aSpottedShape[3], aEntity[2], aEntity[3], 0.0, 0.0)
                        # --------------
                        if not aResult is None: # intersection ?
                            self.setBorder(bReverse, aMouseRelPos, [{'D':aResult[0], 'X':aResult[2], 'Y':aResult[3]}])
                    # ------------------------------------------------------------------------------------------
                    elif aEntity[1] in ("CircleShape", "EllipseShape", "ArcCircleShape", "ArcEllipseShape"):
                        aResult = self.xCompute.getInterLineEllipse2(self.aSpottedShape[2], self.aSpottedShape[3], *aEntity[2:5])
                        # --------------
                        if not aResult is None: # intersection ?
                            self.setBorder(bReverse, aMouseRelPos, [{'D':aResult[i], 'X':aResult[2*i+2], 'Y':aResult[2*i+3]} for i in (0, 1)])
                # ------------------------------------------------------------------------------------------------------
                elif self.aSpottedShape[1] in ("CircleShape", "EllipseShape", "ArcCircleShape", "ArcEllipseShape"): # spotted entity is
                    # ---------------------------------------
                    if aEntity[1] == "LineShape": # border candidate is a line
                        aResult = self.xCompute.getInterLineEllipse2(aEntity[2], aEntity[3], *self.aSpottedShape[2:5])
                        # --------------
                        if not aResult is None: # intersection ?
                            nDist = [{'D':0, 'X':aResult[2], 'Y':aResult[3]}, {'D':0, 'X':aResult[4], 'Y':aResult[5]}]
                            nDist[0]['D'] = self.xCompute.getDist2Arc(*(self.aSpottedShape[2:7]), aCoordB=nDist[0])[0]
                            nDist[1]['D'] = self.xCompute.getDist2Arc(*(self.aSpottedShape[2:7]), aCoordB=nDist[1])[0]
                            self.setBorder(bReverse, aMouseRelPos, nDist)
                    # -------------------------------------------------------------------------------------------
                    elif aEntity[1] in ("CircleShape", "EllipseShape", "ArcCircleShape", "ArcEllipseShape"):
                        aResult = self.xCompute.getInterEllipses(*(self.aSpottedShape[2:5] + aEntity[2:5]))
                        if not aResult is None:
                            nDist = []
                            # --------------
                            if len(aResult) > 0: # intersection ?
                                nDist.append({'X':aResult[0], 'Y':aResult[1]})
                                nDist[0]['D'] = self.xCompute.getDist2Arc(*(self.aSpottedShape[2:7]), aCoordB=nDist[0])[0]
                            # --------------
                            if len(aResult) > 2: # intersection ?
                                nDist.append({'X':aResult[2], 'Y':aResult[3]})
                                nDist[1]['D'] = self.xCompute.getDist2Arc(*(self.aSpottedShape[2:7]), aCoordB=nDist[1])[0]
                            # --------------
                            if len(aResult) > 4: # intersection ?
                                nDist.append({'X':aResult[4], 'Y':aResult[5]})
                                nDist[2]['D'] = self.xCompute.getDist2Arc(*(self.aSpottedShape[2:7]), aCoordB=nDist[2])[0]
                            # --------------
                            if len(aResult) > 6: # intersection ?
                                nDist.append({'X':aResult[6], 'Y':aResult[7]})
                                nDist[3]['D'] = self.xCompute.getDist2Arc(*(self.aSpottedShape[2:7]), aCoordB=nDist[3])[0]
                            self.setBorder(bReverse, aMouseRelPos, nDist)
        # =============
        if bReverse[0]:
            # ------------------------------------------------------------
            if self.aBorderList['IOB']['DELTA'] < 0 and self.aBorderList['IEB']['DELTA'] >= 0:
                self.aBorderList['IOB']['DELTA'] = self.aReverseBorderList['IEB']['DELTA']
                self.aBorderList['IOB']['RATE'] = self.aReverseBorderList['IEB']['RATE']
                self.aBorderList['IOB']["X"] = self.aReverseBorderList['IEB']["X"]
                self.aBorderList['IOB']["Y"] = self.aReverseBorderList['IEB']["Y"]
            # --------------------------------------------------------------
            elif self.aBorderList['IOB']['DELTA'] >= 0 and self.aBorderList['IEB']['DELTA'] < 0:
                self.aBorderList['IEB']['DELTA'] = self.aReverseBorderList['IOB']['DELTA']
                self.aBorderList['IEB']['RATE'] = self.aReverseBorderList['IOB']['RATE']
                self.aBorderList['IEB']["X"] = self.aReverseBorderList['IOB']["X"]
                self.aBorderList['IEB']["Y"] = self.aReverseBorderList['IOB']["Y"]
        # =============
        if bReverse[1]:
            # ------------------------------------------------------------
            if self.aBorderList['OOB']['DELTA'] < 0 and self.aBorderList['OEB']['DELTA'] >= 0:
                self.aBorderList['OOB']['DELTA'] = self.aReverseBorderList['OEB']['DELTA']
                self.aBorderList['OOB']['RATE'] = self.aReverseBorderList['OEB']['RATE']
                self.aBorderList['OOB']["X"] = self.aReverseBorderList['OEB']["X"]
                self.aBorderList['OOB']["Y"] = self.aReverseBorderList['OEB']["Y"]
            # --------------------------------------------------------------
            elif self.aBorderList['OOB']['DELTA'] >= 0 and self.aBorderList['OEB']['DELTA'] < 0:
                self.aBorderList['OEB']['DELTA'] = self.aReverseBorderList['OOB']['DELTA']
                self.aBorderList['OEB']['RATE'] = self.aReverseBorderList['OOB']['RATE']
                self.aBorderList['OEB']["X"] = self.aReverseBorderList['OOB']["X"]
                self.aBorderList['OEB']["Y"] = self.aReverseBorderList['OOB']["Y"]
        return
    # sets the border list
    def setBorder(self, bReverse, aMouseRelPos, nDistBorder):
        if bReverse[0] or bReverse[1]:
            nFullLength = long(round(math.sqrt(self.aSpottedShape[3]['Width'] * self.aSpottedShape[3]['Height']) * 2 * math.pi))
        for nDB in nDistBorder:
            nDelta =  nDB['D'] - aMouseRelPos[0]
            nRate = abs(long(aMouseRelPos[0] * 100 / aMouseRelPos[2]))
            # -----------------------------------------------------------------
            if (nDB['D'] > 5 and nDB['D'] < aMouseRelPos[2]-5) or bReverse[0]: # within segment limits ?
                # -------------
                if nDelta >= 0: # is this a in end border ?
                    # ----------------------------
                    if self.aBorderList['IEB']['DELTA'] < 0: # not yet initialized ?
                        self.aBorderList['IEB']['DELTA'] = abs(nDelta)
                    # --------------------------------------------
                    if abs(nDelta) <= self.aBorderList['IEB']['DELTA']:
                        self.aBorderList['IEB']['DELTA'] = abs(nDelta)
                        self.aBorderList['IEB']['RATE'] =  nRate
                        self.aBorderList['IEB']["X"] = nDB['X']
                        self.aBorderList['IEB']["Y"] = nDB['Y']
                    if bReverse[0]:
                        nReverseDelta = nDelta - nFullLength
                        if self.aReverseBorderList['IEB']['DELTA'] < 0:
                            self.aReverseBorderList['IEB']['DELTA'] = abs(nReverseDelta)
                        if abs(nReverseDelta) <= self.aReverseBorderList['IEB']['DELTA']:
                            self.aReverseBorderList['IEB']['DELTA'] = abs(nReverseDelta)
                            self.aReverseBorderList['IEB']['RATE'] = nRate
                            self.aReverseBorderList['IEB']["X"] = nDB['X']
                            self.aReverseBorderList['IEB']["Y"] = nDB['Y']
                # --------------
                else: # is this a in origin Border ?
                    # ----------------------------
                    if self.aBorderList['IOB']['DELTA'] < 0: # not yet initialized ?
                        self.aBorderList['IOB']['DELTA'] = abs(nDelta)
                    # --------------------------------------------
                    if abs(nDelta) <= self.aBorderList['IOB']['DELTA']:
                        self.aBorderList['IOB']['DELTA'], self.aBorderList['IOB']['RATE'], self.aBorderList['IOB']["X"], self.aBorderList['IOB']["Y"] = abs(nDelta), nRate, nDB['X'], nDB['Y']
                    if bReverse[0]:
                        nReverseDelta = nFullLength + nDelta
                        if self.aReverseBorderList['IOB']['DELTA'] < 0:
                            self.aReverseBorderList['IOB']['DELTA'] = abs(nReverseDelta)
                        if abs(nReverseDelta) <= self.aReverseBorderList['IOB']['DELTA']:
                            self.aReverseBorderList['IOB']['DELTA'] = abs(nReverseDelta)
                            self.aReverseBorderList['IOB']['RATE'] = nRate
                            self.aReverseBorderList['IOB']["X"] = nDB['X']
                            self.aReverseBorderList['IOB']["Y"] = nDB['Y']
            # -----------------
            elif nDB['D'] < -5: # is this the out origin border ?
                # ----------------------------
                if self.aBorderList['OOB']['DELTA'] < 0: # not yet initialized ?
                    self.aBorderList['OOB']['DELTA'] = abs(nDelta)
                if abs(nDelta) <= self.aBorderList['OOB']['DELTA']:
                    self.aBorderList['OOB']['DELTA'], self.aBorderList['OOB']['RATE'], self.aBorderList['OOB']["X"], self.aBorderList['OOB']["Y"] = abs(nDelta), nRate, nDB['X'], nDB['Y']
                if bReverse[1]:
                    nReverseDelta = nFullLength + nDB['D']
                    if self.aReverseBorderList['OOB']['DELTA'] < 0:
                        self.aReverseBorderList['OOB']['DELTA'] = abs(nReverseDelta)
                    if abs(nReverseDelta) <= self.aReverseBorderList['OOB']['DELTA']:
                        self.aReverseBorderList['OOB']['DELTA'] = abs(nReverseDelta)
                        self.aReverseBorderList['OOB']['RATE'] = nRate
                        self.aReverseBorderList['OOB']["X"] = nDB['X']
                        self.aReverseBorderList['OOB']["Y"] = nDB['Y']
            # ---
            elif nDB['D'] > aMouseRelPos[2]+5: # is this the out end border ?
                # ----------------------------
                if self.aBorderList['OEB']['DELTA'] < 0: # not yet initialized ?
                    self.aBorderList['OEB']['DELTA'] = abs(nDelta)
                if abs(nDelta) <= self.aBorderList['OEB']['DELTA']:
                    self.aBorderList['OEB']['DELTA'], self.aBorderList['OEB']['RATE'], self.aBorderList['OEB']["X"], self.aBorderList['OEB']["Y"] = abs(nDelta), nRate, nDB['X'], nDB['Y']
                if bReverse[1]:
                    nReverseDelta = nFullLength - nDB['D']
                    if self.aReverseBorderList['OEB']['DELTA'] < 0:
                        self.aReverseBorderList['OEB']['DELTA'] = abs(nReverseDelta)
                    if abs(nReverseDelta) <= self.aReverseBorderList['OEB']['DELTA']:
                        self.aReverseBorderList['OEB']['DELTA'] = abs(nReverseDelta)
                        self.aReverseBorderList['OEB']['RATE'] = nRate
                        self.aReverseBorderList['OEB']["X"] = nDB['X']
                        self.aReverseBorderList['OEB']["Y"] = nDB['Y']
        return
    # decompose a shape in segments
    def splitShape(self, aShape, aType, nIndx):
        aSegments = []
        aCoord = ({"X": 0,"Y": 0}, {"X": 0,"Y": 0})
        aCenterX = aCenterY = nNbPoint = 0L
        # -------------------------------------------------------------
        if aType in ("LineShape", "PolyLineShape", "PolyPolygonShape"):
            # --------------------------------------
            for i in range(len(aShape.PolyPolygon)):
                # -------------------------------------------
                for j in range(len(aShape.PolyPolygon[i])-1):
                    aSegments.append([tuple(nIndx), "LineShape", {"X": 0,"Y": 0}, {"X": 0,"Y": 0}])
                    aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aShape.PolyPolygon[i][j].X, aShape.PolyPolygon[i][j].Y
                    aSegments[-1][3]["X"], aSegments[-1][3]["Y"] = aShape.PolyPolygon[i][j+1].X, aShape.PolyPolygon[i][j+1].Y
                    aCenterX, aCenterY = aCenterX + aShape.PolyPolygon[i][j].X, aCenterY + aShape.PolyPolygon[i][j].Y
                    nNbPoint = nNbPoint + 1
                    if aType in ("PolyLineShape", "PolyPolygonShape"):
                        aSegments.append([tuple(nIndx), "Centre%s" % aType, {"X": 0,"Y": 0}])
                        aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = (aSegments[-2][2]["X"] + aSegments[-2][3]["X"])/2, (aSegments[-2][2]["Y"] + aSegments[-2][3]["Y"])/2
                if aShape.PolyPolygon[i][0].X != aShape.PolyPolygon[i][-1].X and aShape.PolyPolygon[i][0].Y != aShape.PolyPolygon[i][-1].Y:
                    aCenterX, aCenterY = aCenterX + aShape.PolyPolygon[i][-1].X, aCenterY + aShape.PolyPolygon[i][-1].Y
                    nNbPoint = nNbPoint + 1
            aCenterX, aCenterY = long(round(float(aCenterX) / nNbPoint)), long(round(float(aCenterY) / nNbPoint))
            aSegments.append([tuple(nIndx), "Centre%s" % aType, {"X": 0,"Y": 0}])
            aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCenterX, aCenterY
        # --------------------------------
        elif aType in ("RectangleShape",):
            aSize = aShape.getSize()
            aULCCoord = aShape.getPosition()
            fAngle1 = aShape.RotateAngle * math.pi / 18000
            fCosW = aSize.Width * math.cos(fAngle1)
            fSinW = -aSize.Width * math.sin(fAngle1)
            fCosH = aSize.Height * math.cos(fAngle1 - math.pi / 2)
            fSinH = -aSize.Height * math.sin(fAngle1 - math.pi / 2)
            fCosCW = aShape.CornerRadius * math.cos(fAngle1)
            fSinCW = -aShape.CornerRadius * math.sin(fAngle1)
            fCosCH = aShape.CornerRadius * math.cos(fAngle1 - math.pi / 2)
            fSinCH = -aShape.CornerRadius * math.sin(fAngle1 - math.pi / 2)
            # Vertices 1 - 2
            aCoord[0]["X"], aCoord[0]["Y"] = long(round(aULCCoord.X + fCosCW)), long(round(aULCCoord.Y + fSinCW))
            aCoord[1]["X"], aCoord[1]["Y"] = long(round(aULCCoord.X + fCosW - fCosCW)), long(round(aULCCoord.Y + fSinW - fSinCW))
            aSegments.append([tuple(nIndx), "LineShape", {"X": 0,"Y": 0}, {"X": 0,"Y": 0}])
            aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
            aSegments[-1][3]["X"], aSegments[-1][3]["Y"] = aCoord[1]["X"], aCoord[1]["Y"]
            aSegments.append([tuple(nIndx), "Centre%s" % aType, {"X": 0,"Y": 0}])
            aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = (aSegments[-2][2]["X"] + aSegments[-2][3]["X"])/2, (aSegments[-2][2]["Y"] + aSegments[-2][3]["Y"])/2
            aCenterX, aCenterY = aCenterX + aCoord[0]["X"] + aCoord[1]["X"], aCenterY + aCoord[0]["Y"] + aCoord[1]["Y"]
            nNbPoint = nNbPoint + 2
            # ---------------------
            if aShape.CornerRadius:
                # Corner 2
                aCoord[0]["X"], aCoord[0]["Y"] = long(round(aULCCoord.X + fCosW - fCosCW + fCosCH)), long(round(aULCCoord.Y + fSinW - fSinCW + fSinCH))
                aSegments.append([tuple(nIndx), "ArcCircleShape", {"X": 0,"Y": 0}, {"Width": 0,"Height": 0}, float(0), fAngle1, fAngle1 + math.pi/2])
                aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
                aSegments[-1][3]["Width"] = aSegments[-1][3]["Height"] = aShape.CornerRadius
                aSegments.append((tuple(nIndx), "CentreArcCircleShape", {"X": 0,"Y": 0}))
                aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
                aSegments.append([tuple(nIndx), "CentreArcCircleShape", {"X": 0,"Y": 0}])
                aSegments[-1][2]["X"] = aCoord[0]["X"] + long(round(aShape.CornerRadius * math.cos(fAngle1+math.pi/4)))
                aSegments[-1][2]["Y"] = aCoord[0]["Y"] + long(round(aShape.CornerRadius * math.sin(-(fAngle1+math.pi/4))))
            # Vertices 3 - 4
            aCoord[0]["X"], aCoord[0]["Y"] = long(round(aULCCoord.X + fCosW + fCosCH)), long(round(aULCCoord.Y + fSinW + fSinCH))
            aCoord[1]["X"], aCoord[1]["Y"] = long(round(aULCCoord.X + fCosW + fCosH - fCosCH)), long(round(aULCCoord.Y + fSinW + fSinH - fSinCH))
            aSegments.append([tuple(nIndx), "LineShape", {"X": 0,"Y": 0}, {"X": 0,"Y": 0}])
            aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
            aSegments[-1][3]["X"], aSegments[-1][3]["Y"] = aCoord[1]["X"], aCoord[1]["Y"]
            aSegments.append([tuple(nIndx), "Centre%s" % aType, {"X": 0,"Y": 0}])
            aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = (aSegments[-2][2]["X"] + aSegments[-2][3]["X"])/2, (aSegments[-2][2]["Y"] + aSegments[-2][3]["Y"])/2
            aCenterX, aCenterY = aCenterX + aCoord[0]["X"] + aCoord[1]["X"], aCenterY + aCoord[0]["Y"] + aCoord[1]["Y"]
            nNbPoint = nNbPoint + 2
            # ---------------------
            if aShape.CornerRadius:
                # Corner 3
                aCoord[0]["X"], aCoord[0]["Y"] = long(round(aULCCoord.X + fCosW + fCosH - fCosCW - fCosCH)), long(round(aULCCoord.Y + fSinW + fSinH - fSinCW - fSinCH))
                aSegments.append([tuple(nIndx), "ArcCircleShape", {"X": 0,"Y": 0}, {"Width": 0,"Height": 0}, float(0), fAngle1 + 3*math.pi/2, fAngle1])
                aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
                aSegments[-1][3]["Width"] = aSegments[-1][3]["Height"] = aShape.CornerRadius
                aSegments.append([tuple(nIndx), "CentreArcCircleShape", {"X": 0,"Y": 0}])
                aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
                aSegments.append([tuple(nIndx), "CentreArcCircleShape", {"X": 0,"Y": 0}])
                aSegments[-1][2]["X"] = aCoord[0]["X"] + long(round(aShape.CornerRadius * math.cos(fAngle1-math.pi/4)))
                aSegments[-1][2]["Y"] = aCoord[0]["Y"] + long(round(aShape.CornerRadius * math.sin(-(fAngle1-math.pi/4))))
            # Vertices 5 - 6
            aCoord[0]["X"], aCoord[0]["Y"] = long(round(aULCCoord.X + fCosW + fCosH - fCosCW)), long(round(aULCCoord.Y + fSinW + fSinH - fSinCW))
            aCoord[1]["X"], aCoord[1]["Y"] = long(round(aULCCoord.X + fCosH + fCosCW)), long(round(aULCCoord.Y + fSinH + fSinCW))
            aSegments.append([tuple(nIndx), "LineShape", {"X": 0,"Y": 0}, {"X": 0,"Y": 0}])
            aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
            aSegments[-1][3]["X"], aSegments[-1][3]["Y"] = aCoord[1]["X"], aCoord[1]["Y"]
            aSegments.append([tuple(nIndx), "Centre%s" % aType, {"X": 0,"Y": 0}])
            aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = (aSegments[-2][2]["X"] + aSegments[-2][3]["X"])/2, (aSegments[-2][2]["Y"] + aSegments[-2][3]["Y"])/2
            aCenterX, aCenterY = aCenterX + aCoord[0]["X"] + aCoord[1]["X"], aCenterY + aCoord[0]["Y"] + aCoord[1]["Y"]
            nNbPoint = nNbPoint + 2
            # ---------------------
            if aShape.CornerRadius:
                # Corner 4
                aCoord[0]["X"], aCoord[0]["Y"] = long(round(aULCCoord.X + fCosH + fCosCW - fCosCH)), long(round(aULCCoord.Y + fSinH + fSinCW - fSinCH))
                aSegments.append([tuple(nIndx), "ArcCircleShape", {"X": 0,"Y": 0}, {"Width": 0,"Height": 0}, float(0), fAngle1 + math.pi, fAngle1 + 3*math.pi/2])
                aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
                aSegments[-1][3]["Width"] = aSegments[-1][3]["Height"] = aShape.CornerRadius
                aSegments.append([tuple(nIndx), "CentreArcCircleShape", {"X": 0,"Y": 0}])
                aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
                aSegments.append([tuple(nIndx), "CentreArcCircleShape", {"X": 0,"Y": 0}])
                aSegments[-1][2]["X"] = aCoord[0]["X"] + long(round(aShape.CornerRadius * math.cos(fAngle1-3*math.pi/4)))
                aSegments[-1][2]["Y"] = aCoord[0]["Y"] + long(round(aShape.CornerRadius * math.sin(-(fAngle1-3*math.pi/4))))
            # Vertices 7 - 8
            aCoord[0]["X"], aCoord[0]["Y"] = long(round(aULCCoord.X + fCosH - fCosCH)), long(round(aULCCoord.Y + fSinH - fSinCH))
            aCoord[1]["X"], aCoord[1]["Y"] = long(round(aULCCoord.X + fCosCH)), long(round(aULCCoord.Y + fSinCH))
            aSegments.append([tuple(nIndx), "LineShape", {"X": 0,"Y": 0}, {"X": 0,"Y": 0}])
            aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
            aSegments[-1][3]["X"], aSegments[-1][3]["Y"] = aCoord[1]["X"], aCoord[1]["Y"]
            aSegments.append([tuple(nIndx), "Centre%s" % aType, {"X": 0,"Y": 0}])
            aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = (aSegments[-2][2]["X"] + aSegments[-2][3]["X"])/2, (aSegments[-2][2]["Y"] + aSegments[-2][3]["Y"])/2
            aCenterX, aCenterY = aCenterX + aCoord[0]["X"] + aCoord[1]["X"], aCenterY + aCoord[0]["Y"] + aCoord[1]["Y"]
            nNbPoint = nNbPoint + 2
            # ---------------------
            if aShape.CornerRadius:
                # Corner 1
                aCoord[0]["X"], aCoord[0]["Y"] = long(round(aULCCoord.X + fCosCW + fCosCH)), long(round(aULCCoord.Y + fSinCW + fSinCH))
                aSegments.append([tuple(nIndx), "ArcCircleShape", {"X": 0,"Y": 0}, {"Width": 0,"Height": 0}, float(0), fAngle1 + math.pi/2, fAngle1 + math.pi])
                aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
                aSegments[-1][3]["Width"] = aSegments[-1][3]["Height"] = aShape.CornerRadius
                aSegments.append([tuple(nIndx), "CentreArcCircleShape", {"X": 0,"Y": 0}])
                aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
                aSegments.append([tuple(nIndx), "CentreArcCircleShape", {"X": 0,"Y": 0}])
                aSegments[-1][2]["X"] = aCoord[0]["X"] + long(round(aShape.CornerRadius * math.cos(fAngle1+3*math.pi/4)))
                aSegments[-1][2]["Y"] = aCoord[0]["Y"] + long(round(aShape.CornerRadius * math.sin(-(fAngle1+3*math.pi/4))))
            aCenterX, aCenterY = long(round(float(aCenterX) / nNbPoint)), long(round(float(aCenterY) / nNbPoint))
            aSegments.append([tuple(nIndx), "Centre%s" % aType, {"X": 0,"Y": 0}])
            aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCenterX, aCenterY
        # ------------------------------
        elif aType in ("EllipseShape",):
            aSize = aShape.getSize()
            aULCCoord = aShape.getPosition()
            fAngle1 = aShape.RotateAngle * math.pi / 18000
            fCosW = aSize.Width * math.cos(fAngle1) / 2
            fSinW = -aSize.Width * math.sin(fAngle1) / 2
            fCosH = aSize.Height * math.cos(fAngle1) / 2
            fSinH = -aSize.Height * math.sin(fAngle1) / 2
            aCoord[0]["X"] = long(round(fCosW - fSinH)) + aULCCoord.X
            aCoord[0]["Y"] = long(round(fSinW + fCosH)) + aULCCoord.Y
            # --------------------------------------
            if abs(aSize.Width - aSize.Height) < 10:
                aCadoooType = "CircleShape"
            else:
                aCadoooType = "EllipseShape"
            aSegments.append([tuple(nIndx), "Centre%s" % aCadoooType, {"X": 0,"Y": 0}])
            aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord[0]["X"], aCoord[0]["Y"]
            # --------------------------
            if aShape.CircleKind in (ARC, CUT, SECTION):
                aCadoooType = "Arc" + aCadoooType
            aSegments.append([tuple(nIndx), aCadoooType, {"X": 0,"Y": 0}, {"Width": 0,"Height": 0}, fAngle1, float(aShape.CircleStartAngle)*math.pi/18000, float(aShape.CircleEndAngle)*math.pi/18000])
            aSegments[-1][2]["X"] = aCoord[0]["X"]
            aSegments[-1][2]["Y"] = aCoord[0]["Y"]
            aSegments[-1][3]["Width"], aSegments[-1][3]["Height"] = aSize.Width / 2 , aSize.Height / 2
            aCoord = dict(self.xCompute.getPointOnEllipse(None, (((aSegments[-1][5]>aSegments[-1][6])*2*math.pi + aSegments[-1][6]) - aSegments[-1][5]) / 2.0 + aSegments[-1][5], *aSegments[-1][2:5])[0])
            aSegments.append([tuple(nIndx), "Centre%s" % aCadoooType, {"X": 0,"Y": 0}])
            aSegments[-1][2]["X"], aSegments[-1][2]["Y"] = aCoord["X"], aCoord["Y"]
        return aSegments
    # ******************************************
    # set entity env and get coordinates
    # ******************************************
    def setEntityEnvAndCoord(self, sMode=None, sType=None, spottedShape=None, areaShape=None):
        if sType is None:
            if not spottedShape is None:
                sType = spottedShape[1]
            else:
                return None
        if spottedShape is None and areaShape is None :
            return None
        if not sMode is None:
            if not hasattr(self, "aEnv"):
                self.aEnv = {}
            self.aEnv['MODE'] = [sMode]
        self.aEntityEnvAndCoord['COORDS'] = [[]]
        # ...........................................
        if sType in ('LineShape',):
            self.aEntityEnvAndCoord['UNOTYPE'] = "LineShape"
            self.aEntityEnvAndCoord['ENTITYPE'] = "LINE"
            self.aEntityEnvAndCoord['STEP'] = 1
            self.aEntityEnvAndCoord['MODE'] = [self.aEnv['MODE'][0], "VERTEX"]
            self.aEntityEnvAndCoord['COORDS'][0].extend([{'X': 0, 'Y': 0} for i in range(2)])
            if not spottedShape is None:
                self.aEntityEnvAndCoord['COORDS'][0][0]['X'], self.aEntityEnvAndCoord['COORDS'][0][0]['Y'] = spottedShape[2]['X'], spottedShape[2]['Y']
                self.aEntityEnvAndCoord['COORDS'][0][1]['X'], self.aEntityEnvAndCoord['COORDS'][0][1]['Y'] = spottedShape[3]['X'], spottedShape[3]['Y']
            else:
                self.aEntityEnvAndCoord['COORDS'][0][0]['X'], self.aEntityEnvAndCoord['COORDS'][0][0]['Y'] = areaShape['XA'], areaShape['YA']
                self.aEntityEnvAndCoord['COORDS'][0][1]['X'], self.aEntityEnvAndCoord['COORDS'][0][1]['Y'] = areaShape['XB'], areaShape['YB']
        # ........................................
        elif sType in ('CircleShape',):
            self.aEntityEnvAndCoord['CIRCLEKIND'] = FULL
            self.aEntityEnvAndCoord['UNOTYPE'] = "EllipseShape"
            self.aEntityEnvAndCoord['ENTITYPE'] = "CIRCLE"
            self.aEntityEnvAndCoord['STEP'] = 1
            self.aEntityEnvAndCoord['MODE'] = [self.aEnv['MODE'][0], "RADIUS"]
            self.aEntityEnvAndCoord['COORDS'][0].extend([{'X': 0, 'Y': 0} for i in range(2)])
            if not spottedShape is None:
                self.aEntityEnvAndCoord['COORDS'][0][0]['X'] = spottedShape[2]['X']
                self.aEntityEnvAndCoord['COORDS'][0][0]['Y'] = spottedShape[2]['Y']
                aPoint = self.xCompute.getPointOnEllipse( None, math.pi/4, *spottedShape[2:5])[0]
                self.aEntityEnvAndCoord['COORDS'][0][1]['X'] = aPoint['X']
                self.aEntityEnvAndCoord['COORDS'][0][1]['Y'] = aPoint['Y']
            else:
                self.aEntityEnvAndCoord['COORDS'][0][0]['X'] = areaShape['XC']
                self.aEntityEnvAndCoord['COORDS'][0][0]['Y'] = areaShape['YC']
                self.aEntityEnvAndCoord['COORDS'][0][1]['X'] = areaShape['XR1']
                self.aEntityEnvAndCoord['COORDS'][0][1]['Y'] = areaShape['YR1']
        # .........................................................
        elif sType in ("ArcCircleShape",):
            self.aEntityEnvAndCoord['CIRCLEKIND'] = ARC
            self.aEntityEnvAndCoord['UNOTYPE'] = "EllipseShape"
            self.aEntityEnvAndCoord['ENTITYPE'] = "ARC"
            self.aEntityEnvAndCoord['STEP'] = 3
            self.aEntityEnvAndCoord['MODE'] = [self.aEnv['MODE'][0], "RADIUS", "STARTANGLE", "ENDANGLE"]
            self.aEntityEnvAndCoord['COORDS'][0].extend([{'X': 0, 'Y': 0} for i in range(4)])
            if not spottedShape is None:
                self.aEntityEnvAndCoord['COORDS'][0][0]['X'] = spottedShape[2]['X']
                self.aEntityEnvAndCoord['COORDS'][0][0]['Y'] = spottedShape[2]['Y']
                aPoint = self.xCompute.getPointOnEllipse( None, math.pi/4, *spottedShape[2:5])[0]
                self.aEntityEnvAndCoord['COORDS'][0][1]['X'] = aPoint['X']
                self.aEntityEnvAndCoord['COORDS'][0][1]['Y'] = aPoint['Y']
                aPoint = self.xCompute.getPointOnEllipse( None, spottedShape[5], *spottedShape[2:5])[0]
                self.aEntityEnvAndCoord['COORDS'][0][2]['X'] = aPoint['X']
                self.aEntityEnvAndCoord['COORDS'][0][2]['Y'] = aPoint['Y']
                aPoint = self.xCompute.getPointOnEllipse( None, spottedShape[6], *spottedShape[2:5])[0]
                self.aEntityEnvAndCoord['COORDS'][0][3]['X'] = aPoint['X']
                self.aEntityEnvAndCoord['COORDS'][0][3]['Y'] = aPoint['Y']
            else:
                self.aEntityEnvAndCoord['COORDS'][0][0]['X'] = areaShape['XC']
                self.aEntityEnvAndCoord['COORDS'][0][0]['Y'] = areaShape['YC']
                self.aEntityEnvAndCoord['COORDS'][0][1]['X'] = areaShape['XR1']
                self.aEntityEnvAndCoord['COORDS'][0][1]['Y'] = areaShape['YR1']
                if areaShape['SA'] == "A":
                    self.aEntityEnvAndCoord['COORDS'][0][2]['X'] = areaShape['XA']
                    self.aEntityEnvAndCoord['COORDS'][0][2]['Y'] = areaShape['YA']
                    self.aEntityEnvAndCoord['COORDS'][0][3]['X'] = areaShape['XB']
                    self.aEntityEnvAndCoord['COORDS'][0][3]['Y'] = areaShape['YB']
                else:
                    self.aEntityEnvAndCoord['COORDS'][0][2]['X'] = areaShape['XB']
                    self.aEntityEnvAndCoord['COORDS'][0][2]['Y'] = areaShape['YB']
                    self.aEntityEnvAndCoord['COORDS'][0][3]['X'] = areaShape['XA']
                    self.aEntityEnvAndCoord['COORDS'][0][3]['Y'] = areaShape['YA']
        # ..........................................
        elif sType in ("EllipseShape",):
            self.aEntityEnvAndCoord['CIRCLEKIND'] = FULL
            self.aEntityEnvAndCoord['UNOTYPE'] = "EllipseShape"
            self.aEntityEnvAndCoord['ENTITYPE'] = "ELLIPSE"
            self.aEntityEnvAndCoord['STEP'] = 2
            self.aEntityEnvAndCoord['MODE'] = [self.aEnv['MODE'][0], "RADIUS", "RADIUS2"]
            self.aEntityEnvAndCoord['COORDS'][0].extend([{'X': 0, 'Y': 0} for i in range(3)])
            if not spottedShape is None:
                self.aEntityEnvAndCoord['COORDS'][0][0]['X'] = spottedShape[2]['X']
                self.aEntityEnvAndCoord['COORDS'][0][0]['Y'] = spottedShape[2]['Y']
                aPoint = self.xCompute.getPointOnEllipse( None, 0.0, *spottedShape[2:5])[0]
                self.aEntityEnvAndCoord['COORDS'][0][1]['X'] = aPoint['X']
                self.aEntityEnvAndCoord['COORDS'][0][1]['Y'] = aPoint['Y']
                aPoint = self.xCompute.getPointOnEllipse( None, math.pi/2, *spottedShape[2:5])[0]
                self.aEntityEnvAndCoord['COORDS'][0][2]['X'] = aPoint['X']
                self.aEntityEnvAndCoord['COORDS'][0][2]['Y'] = aPoint['Y']
            else:
                self.aEntityEnvAndCoord['COORDS'][0][0]['X'] = areaShape['XC']
                self.aEntityEnvAndCoord['COORDS'][0][0]['Y'] = areaShape['YC']
                self.aEntityEnvAndCoord['COORDS'][0][1]['X'] = areaShape['XR1']
                self.aEntityEnvAndCoord['COORDS'][0][1]['Y'] = areaShape['YR1']
                self.aEntityEnvAndCoord['COORDS'][0][2]['X'] = areaShape['XR2']
                self.aEntityEnvAndCoord['COORDS'][0][2]['Y'] = areaShape['YR2']
        # .........................................................
        elif sType in ("ArcEllipseShape",):
            self.aEntityEnvAndCoord['CIRCLEKIND'] = ARC
            self.aEntityEnvAndCoord['UNOTYPE'] = "EllipseShape"
            self.aEntityEnvAndCoord['ENTITYPE'] = "ELLARC"
            self.aEntityEnvAndCoord['STEP'] = 4
            self.aEntityEnvAndCoord['MODE'] = [self.aEnv['MODE'][0], "RADIUS", "RADIUS2", "STARTANGLE", "ENDANGLE"]
            self.aEntityEnvAndCoord['COORDS'][0].extend([{'X': 0, 'Y': 0} for i in range(5)])
            if not spottedShape is None:
                self.aEntityEnvAndCoord['COORDS'][0][0]['X'] = spottedShape[2]['X']
                self.aEntityEnvAndCoord['COORDS'][0][0]['Y'] = spottedShape[2]['Y']
                aPoint = self.xCompute.getPointOnEllipse( None, 0.0, *spottedShape[2:5])[0]
                self.aEntityEnvAndCoord['COORDS'][0][1]['X'] = aPoint['X']
                self.aEntityEnvAndCoord['COORDS'][0][1]['Y'] = aPoint['Y']
                aPoint = self.xCompute.getPointOnEllipse( None, math.pi/2, *spottedShape[2:5])[0]
                self.aEntityEnvAndCoord['COORDS'][0][2]['X'] = aPoint['X']
                self.aEntityEnvAndCoord['COORDS'][0][2]['Y'] = aPoint['Y']
                aPoint = self.xCompute.getPointOnEllipse( None, spottedShape[5], *spottedShape[2:5])[0]
                self.aEntityEnvAndCoord['COORDS'][0][3]['X'] = aPoint['X']
                self.aEntityEnvAndCoord['COORDS'][0][3]['Y'] = aPoint['Y']
                aPoint = self.xCompute.getPointOnEllipse( None, spottedShape[6], *spottedShape[2:5])[0]
                self.aEntityEnvAndCoord['COORDS'][0][4]['X'] = aPoint['X']
                self.aEntityEnvAndCoord['COORDS'][0][4]['Y'] = aPoint['Y']
            else:
                self.aEntityEnvAndCoord['COORDS'][0][0]['X'] = areaShape['XC']
                self.aEntityEnvAndCoord['COORDS'][0][0]['Y'] = areaShape['YC']
                self.aEntityEnvAndCoord['COORDS'][0][1]['X'] = areaShape['XR1']
                self.aEntityEnvAndCoord['COORDS'][0][1]['Y'] = areaShape['YR1']
                self.aEntityEnvAndCoord['COORDS'][0][2]['X'] = areaShape['XR2']
                self.aEntityEnvAndCoord['COORDS'][0][2]['Y'] = areaShape['YR2']
                if areaShape['SA'] == "A":
                    self.aEntityEnvAndCoord['COORDS'][0][3]['X'] = areaShape['XA']
                    self.aEntityEnvAndCoord['COORDS'][0][3]['Y'] = areaShape['YA']
                    self.aEntityEnvAndCoord['COORDS'][0][4]['X'] = areaShape['XB']
                    self.aEntityEnvAndCoord['COORDS'][0][4]['Y'] = areaShape['YB']
                else:
                    self.aEntityEnvAndCoord['COORDS'][0][3]['X'] = areaShape['XB']
                    self.aEntityEnvAndCoord['COORDS'][0][3]['Y'] = areaShape['YB']
                    self.aEntityEnvAndCoord['COORDS'][0][4]['X'] = areaShape['XA']
                    self.aEntityEnvAndCoord['COORDS'][0][4]['Y'] = areaShape['YA']
        else:
            self.aEntityEnvAndCoord = {'UNOTYPE':None, 'ENTITYPE':None, 'STEP':0, 'MODE':[self.aEnv['MODE'][0]], 'COORDS':None}
        return 
