# $Id: $
# -*- coding: latin-1 -*-
## ********************************************************************************
## link2obj jeu jan 26 20:18:45 CET 2005
## Copyright (C) 2004-2009 Grard Deneux
## gerard.deneux@free.fr

## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.

## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
## Lesser General Public License for more details.
## http://www.opensource.org/licenses/lgpl-license.php

## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
## ******************************************************************************
try:
    import math
except ImportError:
    print "link2obj : probleme d'import"
# ======================================================================================
# Gerer les liens aux objets lors des traces des entit
# Handling links to object while drawing entities
# ======================================================================================
class Link2Obj():
    """Gere les liens aux objets
    Handles links to objects"""
    def __init__(self, ctx):
        import fr.free.deneux.gerard.cadooo.compute
        import fr.free.deneux.gerard.cadooo.entitiespot
        import fr.free.deneux.gerard.cadooo.magneticpoleshape
        smgr = ctx.ServiceManager
        # usefull coordinates
        self.aMagneticCoord = {'X':0, 'Y':0}
        self.xCompute = fr.free.deneux.gerard.cadooo.compute.Compute(ctx)
        self.xEntitieSpot = fr.free.deneux.gerard.cadooo.entitiespot.EntitieSpot(ctx)
        self.xMagneticPoleShape = fr.free.deneux.gerard.cadooo.magneticpoleshape.MagneticPoleShape(ctx)
        self.bRedraw = False
    # ========================================
    # aNamedValues as tuple of NamedValue
    #   Name         Value
    # SPOTPARM
    # L2OSTATE    tuple of 5 integers (0 or 1)
    # VALIDLIST   integer (0 or 1)
    # ENTITYPE    string shape type
    # VALIDORIGIN integer (0 or 1)
    # ORIGINCOORD tuple of 2 integers
    # VALIDL2O    integer (0 or 1)
    # L2OCOORD    tuple of 2 integers
    # MOUSEEVENT  struct MouseEvent
    # ========================================
    def execute(self, aNamedValues):
        aEnv = {}
        aEnv.update(aNamedValues)
        if self.bRedraw:
            self.xEntitieSpot.bValidList = False
            self.bRedraw = False
        aSpotParm = dict([(k, aEnv[k]) for k in (aEnv['SPOTPARM'] + aEnv['MOUSEPARM'])])
        aEnv.update(self.xEntitieSpot.execute(aSpotParm))
        if self.xEntitieSpot.aSpotList is None or list(aEnv['L2OSTATE']).count(0) == 5:
            aEnv['MAGNETKIND'] = "0"
            aEnv['VALIDL2O'] = 0
            aEnv.update(self.xMagneticPoleShape.execute(aEnv))
            return aEnv
        ## which point is the closest
        aMouseParm = dict([(k, aEnv[k]) for k in aEnv['MOUSEPARM']])
        aMouseCoord = self.xCompute.transformMouseCoord(aMouseParm)
        self.nElectedDist = int(math.sqrt(aMouseCoord['X'] ** 2 + aMouseCoord['Y'] ** 2))
        aCoord = {'X': 0, 'Y': 0}
        aSegments = [list(i[1:]) for i in self.xEntitieSpot.aSpotList]
        ## Magnetic ends
        if aEnv['L2OSTATE'][0]:
            for aSeg in aSegments:
                # --- a type line ---
                if aSeg[0] in ("LineShape",):
                    if self.testCoord((aSeg[1]['X'], aSeg[1]['Y'], aSeg[2]['X'], aSeg[2]['Y']), aMouseCoord):
                        aEnv['MAGNETKIND'] = "e"
                if aSeg[0] in ("ArcCircleShape", "ArcEllipseShape"):
                    fCosW = aSeg[2]['Width'] * math.cos(-aSeg[3])
                    fSinW = aSeg[2]['Width'] * math.sin(-aSeg[3])
                    fCosH = aSeg[2]['Height'] * math.cos(-aSeg[3])
                    fSinH = aSeg[2]['Height'] * math.sin(-aSeg[3])
                    fAngle1 = -aSeg[4]
                    fAngle2 = -aSeg[5]
                    aCoord['X'] = aSeg[1]['X'] + fCosW * math.cos(fAngle1) - fSinH * math.sin(fAngle1)
                    aCoord['Y'] = aSeg[1]['Y'] + fSinW * math.cos(fAngle1) + fCosH * math.sin(fAngle1)
                    if self.testCoord((aCoord['X'], aCoord['Y']), aMouseCoord):
                        aEnv['MAGNETKIND'] = "e"
                    aCoord['X'] = aSeg[1]['X'] + fCosW * math.cos(fAngle2) - fSinH * math.sin(fAngle2)
                    aCoord['Y'] = aSeg[1]['Y'] + fSinW * math.cos(fAngle2) + fCosH * math.sin(fAngle2)
                    if self.testCoord((aCoord['X'], aCoord['Y']), aMouseCoord):
                        aEnv['MAGNETKIND'] = "e"
        ## Magnetic intersection
        if aEnv['L2OSTATE'][1]:
            for i in range(len(aSegments)-1):
                for j in range(i+1, len(aSegments)):
                    ## intersection line - line
                    if aSegments[i][0] in ("LineShape",) and aSegments[j][0] in ("LineShape",):
                        fAngle1 = self.xCompute.toPolarCoord(aSegments[i][1], aSegments[i][2])[1]
                        fAngle2 = self.xCompute.toPolarCoord(aSegments[j][1], aSegments[j][2])[1]
                        nDistance = self.xCompute.getInterLines(aSegments[i][1], fAngle1, aSegments[j][1], fAngle2)
                        if not nDistance is None:
                            if self.testCoord(tuple(nDistance[2:]), aMouseCoord):
                                aEnv['MAGNETKIND'] = "i"
                    ## intersection line - ellipse
                    elif aSegments[i][0] in ("LineShape",) and aSegments[j][0] in ("CircleShape", "ArcCircleShape", "EllipseShape", "ArcEllipseShape"):
                        fAngle1 = self.xCompute.toPolarCoord(aSegments[i][1], aSegments[i][2])[1]
                        nDistance = self.xCompute.getInterLineEllipse(aSegments[i][1], fAngle1, *aSegments[j][1:4])
                        if not nDistance is None:
                            if self.testCoord(tuple(nDistance[2:]), aMouseCoord):
                                aEnv['MAGNETKIND'] = "i"
                    ## intersection ellipse - line
                    elif aSegments[j][0] in ("LineShape",) and aSegments[i][0] in ("CircleShape", "ArcCircleShape", "EllipseShape", "ArcEllipseShape"):
                        fAngle1 = self.xCompute.toPolarCoord(aSegments[j][1], aSegments[j][2])[1]
                        nDistance = self.xCompute.getInterLineEllipse(aSegments[j][1], fAngle1, *aSegments[i][1:4])
                        if not nDistance is None:
                            if self.testCoord(tuple(nDistance[2:]), aMouseCoord):
                                aEnv['MAGNETKIND'] = "i"
                        # intersection ellipse - ellipse
                    elif not aSegments[i][0] in ("LineShape", ) and not aSegments[j][0] in ("LineShape",) and aSegments[i][0].count("Centre") == 0 and aSegments[j][0].count("Centre") == 0:
                        nDistance = self.xCompute.getInterEllipses(*(aSegments[i][1:4]+ aSegments[j][1:4]))
                        if not nDistance is None:
                            if self.testCoord(tuple(nDistance), aMouseCoord):
                                aEnv['MAGNETKIND'] = "i"
        # Magnetic middle
        if aEnv['L2OSTATE'][2]:
            for aSeg in aSegments:
                if aSeg[0].count("Centre") > 0:
                    if self.testCoord((aSeg[1]['X'], aSeg[1]['Y']), aMouseCoord):
                        aEnv['MAGNETKIND'] = "m"
        # Magnetic perpendicular
        if aEnv['L2OSTATE'][3] and aEnv['VALIDORIGIN'] and aEnv['ENTITYPE'] == "LINE":
            for aSeg in aSegments:
                if aSeg[0] in ("LineShape",):
                    fAngle1 = self.xCompute.toPolarCoord(aSeg[1], aSeg[2])[1]
                    nDistance = self.xCompute.getTangentLineCircle(aSeg[1], fAngle1, aEnv['ORIGINCOORD'])
                    if self.testCoord(tuple(nDistance), aMouseCoord):
                        aEnv['MAGNETKIND'] = "p"
        # Magnetic tangent
        if aEnv['L2OSTATE'][4] and aEnv['VALIDORIGIN']:
            if aEnv['ENTITYPE'] in ("CIRCLE", "ARC"):
                for aSeg in aSegments:
                    if aSeg[0] in ("LineShape",):
                        fAngle1 = -math.atan2(aSeg[2]['Y']-aSeg[1]['Y'], aSeg[2]['X']-aSeg[1]['X']) # self.compute('toPolarCoord', aSeg[1], aSeg[2])[1]
                        nDistance = self.xCompute.getTangentLineCircle(aSeg[1], fAngle1, aEnv['ORIGINCOORD'])
                        if not nDistance is None:
                            if self.testCoord(tuple(nDistance), aMouseCoord):
                                aEnv['MAGNETKIND'] = "t"
                    elif aSeg[0] in ("EllipseShape", "ArcEllipseShape", "CircleShape", "ArcCircleShape"):
                        nDistance = self.xCompute.getTangentEllipseCircle(aSeg[1], aSeg[2], aSeg[3], aEnv['ORIGINCOORD'])
                        if not nDistance is None:
                            if self.testCoord(tuple(nDistance), aMouseCoord):
                                aEnv['MAGNETKIND'] = "t"
            if aEnv['ENTITYPE'] in ("LINE",):
                for aSeg in aSegments:
                    if aSeg[0] in ("EllipseShape", "ArcEllipseShape", "CircleShape", "ArcCircleShape"):
                        nDistance = self.xCompute.getTangentEllipseLine(aSeg[1], aSeg[2], aSeg[3], aEnv['ORIGINCOORD'])
                        if not nDistance is None:
                            if self.testCoord(nDistance, aMouseCoord):
                                aEnv['MAGNETKIND'] = "t"
        if self.nElectedDist <= aEnv['MAGNETSIZE']*aEnv['DEVICEXSCALE']:
            aEnv['L2OCOORD'].update(self.aMagneticCoord)
            aEnv['VALIDL2O'] = 1
        else:
            aEnv['VALIDL2O'] = 0
            aEnv['MAGNETKIND'] = "0"
        aEnv.update(self.xMagneticPoleShape.execute(aEnv))
        return aEnv
    # test and set coordinates of the closest point
    def testCoord(self, nListCoord, aMouseCoord):
        bFlg = False
        for nX, nY in [(nListCoord[i], nListCoord[i+1]) for i in range(0,len(nListCoord),2)]:
            nCandidateDist = long(round(math.hypot(nX-aMouseCoord['X'], nY-aMouseCoord['Y'])))
            if  nCandidateDist <= self.nElectedDist:
                self.nElectedDist = nCandidateDist
                self.aMagneticCoord['X'], self.aMagneticCoord['Y'] = nX, nY
                bFlg = True
        return bFlg
