#
# -*- coding: latin-1 -*-
## ********************************************************************************
## 
## Copyright (C) 2005-2009 Grard Deneux
## gerard.deneux@free.fr

## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.

## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
## Lesser General Public License for more details.
## http://www.opensource.org/licenses/lgpl-license.php

## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
## ******************************************************************************
try:
    import math, traceback
    from types import *
    from com.sun.star.awt.KeyModifier import SHIFT
except ImportError:
    print "probleme d'import"
#===============================
# gestion des evenements clavier
# handling keyboard events
#===============================
class OffsetKeyHandler():
    """gestion des evenements clavier
    handling keyboard events"""
    def __init__(self, ctx):
        from fr.free.deneux.gerard.cadooo.datainputbox import DataInputBox
        from fr.free.deneux.gerard.cadooo.compute import Compute
        from fr.free.deneux.gerard.cadooo.msgl10n import MsgL10n
        self.ctx = ctx
        self.smgr = self.ctx.ServiceManager
        # Datum needed by coordinates input form
        self.aCoordData = {}
        self.aCoordData["TITLE"]  = "No Title"
        self.aCoordData["FLAG"] = 0
        self.aCoordData["COMMENT"] = "No comment"
        self.aCoordData["X"] = 0
        self.aCoordData["XUNIT"] = "m"
        self.aCoordData["Y"] = 0
        self.aCoordData["YUNIT"] = "m"
        self.aCoordData["RETURN"] = False
        self.xCoordInput = DataInputBox(self.ctx)
        self.xCompute = Compute(self.ctx)
        self.xMsgL10n = MsgL10n(self.ctx)
    # ==================================
    # actions clavier - keyboard actions
    # aNamedValues as tuple of NamedValue
    #   Name         Value
    # KEYCODE     integer
    # L2OSTATE    tuple of 5 integers (0 or 1)
    # DRAWSCALE
    # DRAWSCALE
    # CANCELJOB
    # ORTHOMODE
    # SHAPE
    # ==================================
    def execute(self, aArgs):
        aEnv = {}
        aEnv.update(aArgs)
        lKey = aEnv['MAGNETKEY'].keys()
        lKey.sort()
        # ----------------------------------------------------------------------------------------------------
        if aEnv['KEYEVENT']['COD'] in lKey and not aEnv['KEYEVENT']['MOD']:
            # e,   i,   m,   p,   t => link 2 entities
            i = lKey.index(aEnv['KEYEVENT']['COD'])
            bKHState = list(aEnv['L2OSTATE'])
            bKHState[i] = not bKHState[i]
            aEnv['L2OSTATE'] = tuple(bKHState)
            aStatusBarCheckBox = aEnv['STATUSBARCONTROL'][aEnv['KEYEVENT']['COD']]
            if aStatusBarCheckBox.getState() != aEnv['L2OSTATE'][i]:
                aStatusBarCheckBox.setState(aEnv['L2OSTATE'][i])
        # --------------------------------------------------------------------------------------------------
        elif aEnv['KEYEVENT']['COD'] in ('LEN', 'RAD') and not aEnv['KEYEVENT']['MOD']:
            # ..............................l, ..r
            # ---------------------------------------------------------------------------
            aTextList = ("k0", "k1")
            aText = self.xMsgL10n.getText(aTextList)
            self.aCoordData["TITLE"]  = aText[0]
            self.aCoordData["FLAG"] = 0
            self.aCoordData["COMMENT"] = aText[1]
            self.aCoordData["RETURN"] = False
            if aEnv['OFFSET'] is None:
                nOffset = 0
            else:
                nOffset = aEnv['OFFSET']
            self.aCoordData["X"] = self.xCompute.transform2UserUnit(nOffset, aEnv['UNITS'], aEnv['UNITSRATIO'], aEnv['MEASUREUNIT'], aEnv['DRAWSCALE'])
            aInputData = dict([(k , aEnv[k]) for k in aEnv['INPUTDATAPARM']])
            self.aCoordData.update(aInputData)
            self.aCoordData.update(self.xCoordInput.execute(self.aCoordData))
            # ---------------------------
            if self.aCoordData["RETURN"]:
                aEnv['OFFSET'] = long(round(self.aCoordData['X']))
            elif aEnv['OFFSET'] is None:
                aEnv['CANCELJOB'] = True
        # --------------------------------------------------------------------------------
        elif aEnv['KEYEVENT']['COD'] in ('ESC',) and not aEnv['KEYEVENT']['MOD']:
            #                                 Esc => kill job
            aEnv['CANCELJOB'] = True
        # -------------------------------------------------------------------------------
        elif aEnv['KEYEVENT']['COD'] in ('AXOCTRL',) and not aEnv['KEYEVENT']['MOD']:
            #                                  a => axonometric mode
            aEnv['ORTHOMODE'] = not aEnv['ORTHOMODE']
            aStatusBarCheckBox = aEnv['STATUSBARCONTROL']['AXOCTRL']
            if aStatusBarCheckBox.getState() != aEnv['ORTHOMODE']:
                aStatusBarCheckBox.setState(aEnv['ORTHOMODE'])
        # --------------------------------------------------------------------------------
        elif aEnv['KEYEVENT']['COD'] in ('MULTICTRL',) and not aEnv['KEYEVENT']['MOD']:
            #                                  * => repeat mode
            aEnv['REPEATMODE'] = not aEnv['REPEATMODE']
            aStatusBarCheckBox = aEnv['STATUSBARCONTROL']['MULTICTRL']
            if aStatusBarCheckBox.getState() != aEnv['REPEATMODE']:
                aStatusBarCheckBox.setState(aEnv['REPEATMODE'])
        return aEnv
