#	$Id$
# -*- coding: latin-1 -*-
## ********************************************************************************
## stairsdraw dimanche 27 avril 2008, 10:51:39 (UTC+0200)

## Copyright (C) 2008-2009 Gerard Deneux
## gerard.deneux@free.fr

## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.

## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
## Lesser General Public License for more details.
## http://www.opensource.org/licenses/lgpl-license.php

## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
## ******************************************************************************
try:
    import uno, math
    from com.sun.star.drawing.CircleKind import FULL, SECTION, CUT, ARC
except ImportError:
    print "probleme d'import"
#======================================================
# a class to draw an entity
# une classe pour tracer une entite
#======================================================
class StairsDraw():
    """Trace un escalier balanc
    Draws a balanced stairway"""
    def __init__(self, ctx):
        import fr.free.deneux.gerard.cadooo.compute
        import fr.free.deneux.gerard.cadooo.entitiespot
        import fr.free.deneux.gerard.cadooo.messagebox
        import fr.free.deneux.gerard.cadooo.msgl10n
        import fr.free.deneux.gerard.cadooo.entityposize
        self.ctx = ctx
        self.xCompute = fr.free.deneux.gerard.cadooo.compute.Compute( self.ctx)
        self.xEntitySpot = fr.free.deneux.gerard.cadooo.entitiespot.EntitieSpot( self.ctx)
        self.xMessage = fr.free.deneux.gerard.cadooo.messagebox.MessageBox( self.ctx)
        self.xMsgL10n = fr.free.deneux.gerard.cadooo.msgl10n.MsgL10n( self.ctx)
        self.aTranslation = self.xMsgL10n.getText(["s%d" % i for i in range(8)])
        self.xEntityPosSize = fr.free.deneux.gerard.cadooo.entityposize.EntityPosSize( self.ctx)
        self.bRedraw = False
    # ==============================
    def execute(self, aArgs):
        self.aEnv = {}
        self.aEnv.update(aArgs)
        # -------------------------
        # check selection
        # -------------------------
        if self.aEnv['STAGE'] == 0:
            # ++++++++++++++++++++++++++++++++++
            if self.aEnv['CONTOURLIST'] is None:
                nNbSeg = 0
                self.aEnv['STAGE'] = 1
                return self.aEnv
            self.aEnv['CONTOURLIST'] = [c for c in self.aEnv['CONTOURLIST'] if not c[1] in ('CircleShape', 'ArcCircleShape', 'EllipseShape', 'ArcEllipseShape')]
            nNbSeg = len(self.aEnv['CONTOURLIST'])
            # ++++++++++++
            if nNbSeg > 8:
                self.xMessage.execute(self.aTranslation[2], "errorbox")
                self.aEnv['CANCELJOB'] = True
                return self.aEnv
            self.aEnv['STAGE'] = 1
            return self.aEnv
        # ------------------------------
        # select contour or first rise
        # ------------------------------
        if self.aEnv['STAGE'] in (1, 3):
            if self.aEnv['CONTOURLIST'] is None:
                self.aEnv['CONTOURLIST'] = []
            nNbSeg = [len(self.aEnv['CONTOURLIST']), None, 1][self.aEnv['STAGE'] - 1]
            sText = "%s %s %d" % (self.aEnv['COMMENT'][0], self.aEnv['COMMENT'][self.aEnv['STAGE']], nNbSeg)
            self.aEnv['STATUSBARCONTROL']['INFOCTRL'].setText(sText)
            if self.bRedraw:
                self.xEntitySpot.bValidList = False
                self.bRedraw = False
            self.aEnv['COORDS'] = None
            aSpotParm = self.aEnv.fromkeys(self.aEnv['SPOTPARM'] + self.aEnv['MOUSEPARM'])
            for aKey in aSpotParm.iterkeys():
                aSpotParm[aKey] = self.aEnv[aKey]
            self.aEnv.update(self.xEntitySpot.execute(aSpotParm))
            # +++++++++++++++++++++++++++++++++++++++
            if not self.xEntitySpot.aSpottedShape is None:
                aNewSpotID = self.xEntitySpot.aSpottedShape[0] + (self.xEntitySpot.aSpottedShape[-1],)
                # ++++++++++++++++++++++++++++++++++++++
                if aNewSpotID == self.aEnv['SPOTTEDID']:
                    self.aEnv['STEP'] = 0
                else:
                    self.removeTempShapes()
                    self.aEnv.update(self.xEntitySpot.aEntityEnvAndCoord)
                    self.aEnv['SPOTTEDID'] = self.xEntitySpot.aSpottedShape[0][:] + (self.xEntitySpot.aSpottedShape[-1],)
            # ++++++++++++++++++++++++++++++++++++++
            elif not self.aEnv['SPOTTEDID'] is None:
                self.aEnv['COORDS'] = None
                self.removeTempShapes()
            if not self.aEnv['COORDS'] is None:
                self.aEnv['SHAPESTATUS'] = "transient"
            else:
                self.aEnv['COORDS'] = None
            return self.aEnv
        # ---------------------------
        # get stairway parameters
        # ---------------------------
        elif self.aEnv['STAGE'] == 2:
            from fr.free.deneux.gerard.cadooo.stairwayinputbox import StairwayInputBox
            sText = "%s %s" % (self.aEnv['COMMENT'][0], self.aEnv['COMMENT'][self.aEnv['STAGE']])
            self.aEnv['STATUSBARCONTROL']['INFOCTRL'].setText(sText)
            xStepsCountDialog = StairwayInputBox(self.ctx)
            aStairwayNamedValues = {}
            aStairwayNamedValues["BUTTONSTATE"] = False
            aStairwayNamedValues["STEPSCOUNT"] = 17
            aStairwayNamedValues["FIRSTSTEPBAL"] = 0
            aStairwayNamedValues["LASTSTEPBAL"] = 0
            aStairwayNamedValues["STRIDELINEDIST"] = self.xCompute.transform2UserUnit( 50000*self.aEnv['DRAWSCALE'], self.aEnv['UNITS'], self.aEnv['UNITSRATIO'], self.aEnv['MEASUREUNIT'], self.aEnv['DRAWSCALE'])
            aStairwayNamedValues["UNITS"] = self.aEnv['UNITS']
            aStairwayNamedValues["MEASUREUNIT"] = self.aEnv['MEASUREUNIT']
            aStairwayNamedValues["UNITSRATIO"] = self.aEnv['UNITSRATIO']
            aStairwayNamedValues["DRAWSCALE"] = self.aEnv['DRAWSCALE']
            aStairwayNamedValues["WINDOW"] = self.aEnv['COMPONENTWINDOW']
            aStairwayNamedValues.update(xStepsCountDialog.execute(aStairwayNamedValues))
            if aStairwayNamedValues["BUTTONSTATE"]:
                self.nStepsCount = aStairwayNamedValues["STEPSCOUNT"]
                self.n1stStepBal = aStairwayNamedValues["FIRSTSTEPBAL"]
                self.nLastStepBal = aStairwayNamedValues["LASTSTEPBAL"]
                self.nStridePos = aStairwayNamedValues["STRIDELINEDIST"]
                self.aEnv['STAGE'] = 3
            else:
                self.xMessage.execute(self.aTranslation[1], "infobox")
                self.aEnv['CANCELJOB'] = True
                return self.aEnv
            del xStepsCountDialog
            del aStairwayNamedValues
            return self.aEnv
        # ---------------------
        # Lists standardisation
        # ---------------------
        elif self.aEnv['STAGE'] == 4:
            sText = "%s %s" % (self.aEnv['COMMENT'][0], self.aEnv['COMMENT'][self.aEnv['STAGE']])
            self.aEnv['STATUSBARCONTROL']['INFOCTRL'].setText(sText)
            # -----------------------------------------------------------------
            # getting vertex coordinates - recupere les coordonnees des sommets
            # -----------------------------------------------------------------
            nEntitySelected = len(self.aEnv['CONTOURLIST'])
            aCoords = []
            aPolars = []
            i, n = (0, 1)
            index = range(nEntitySelected)
            # create vertex coordinates of real contour
            for i in range(nEntitySelected):
                rep = self.xCompute.getInterLines2( self.aEnv['CONTOURLIST'][i-1][2], self.aEnv['CONTOURLIST'][i-1][3], self.aEnv['CONTOURLIST'][i][2], self.aEnv['CONTOURLIST'][i][3], 0 , 0)
                if rep is None:
                    self.xMessage.execute(self.aTranslation[6], "errorbox")
                    self.aEnv['CONTOURLIST'] = None
                    self.aEnv['STAGE'] = 0
                    return self.aEnv
                aCoords.append({'X': 0, 'Y': 0})
                aCoords[-1]['X'] = rep[2]
                aCoords[-1]['Y'] = rep[3]
                if i > 0:
                    aPolars.append(self.xCompute.toPolarCoord( aCoords[-2], aCoords[-1]))
            aPolars.append(self.xCompute.toPolarCoord(aCoords[-1], aCoords[0]))
            nStepWidth = aPolars[0][0]
            #print "\n".join(["X = %d Y = %d" % (c["X"], c["Y"]) for c in aCoords])
            #print "\n".join(["d = %d a = %f" % (c[0], c[1]) for c in aPolars])
            # -------------------------------------------------
            fStridePos = aPolars[1][1] - aPolars[0][1] + math.pi
            if fStridePos > 2 * math.pi:
                fStridePos = fStridePos - 2 * math.pi
            # ecart de la ligne de foulee au collet
            # path line offset from collar
            if self.nStridePos == 0:
                if (nStepWidth / self.aEnv['DRAWSCALE']) < 100000:
                    self.nStridePos = long(nStepWidth / 2 * math.sin(fStridePos))
                else:
                    self.nStridePos = long(65000 * self.aEnv['DRAWSCALE'] * math.sin(fStridePos))
            else:
                self.nStridePos = long(self.nStridePos * (1 - 2 * (math.sin(fStridePos) < 0)))
            # ----------------------
            if nEntitySelected == 6:
                aCoords.insert(5, {'X': 0, 'Y': 0})
                fAngle = (self.xCompute.toPolarCoord( aCoords[6], aCoords[2])[1] + fStridePos) % (2 * math.pi)
                aPolars.insert(5, [0, fAngle])
                aCoords[5]['X'], aCoords[5]['Y'] = aCoords[6]['X'], aCoords[6]['Y']
                fAngle = (fAngle + math.pi) % (2 * math.pi)
                aPolars.insert(2, [0, fAngle])
                aCoords.insert(2, {'X': 0, 'Y': 0})
                aCoords[2]['X'], aCoords[2]['Y'] = aCoords[3]['X'], aCoords[3]['Y']
            # ---------------------    
            if nEntitySelected == 7:
                aCoords.insert(2, {'X': 0, 'Y': 0})
                fAngle = aPolars[1][1] + fStridePos
                aPolars.insert(2, [0, fAngle])
                aCoords[2]['X'], aCoords[2]['Y'] = aCoords[3]['X'], aCoords[3]['Y']
                nIndexPairs[1] = nIndexPairs[1] + 1
            #print "\n".join(["X = %d Y = %d" % (c["X"], c["Y"]) for c in aCoords])
            #print "\n".join(["d = %d a = %f" % (c[0], c[1]) for c in aPolars])
            # print "Drawing path line - trace de la ligne de foulee"
            self.aEnv['GROUPSHAPE'] = self.aEnv['MODEL'].createInstance("com.sun.star.drawing.GroupShape")
            self.aEnv['CONTROLLER'].CurrentPage.add(self.aEnv['GROUPSHAPE'])
            aLineDash = uno.createUnoStruct("com.sun.star.drawing.LineDash")
            aLineDash.Style = uno.Enum("com.sun.star.drawing.DashStyle", "RECT")
            aLineDash.Dots = 1
            aLineDash.DotLen = 100
            aLineDash.Dashes = 1
            aLineDash.DashLen = 500
            aLineDash.Distance = 100
            aCoord = [{'X': 0, 'Y': 0}, {'X': 0, 'Y': 0}, {'X': 0, 'Y': 0}, {'X': 0, 'Y': 0}]
            aNamedValue = (uno.createUnoStruct("com.sun.star.beans.NamedValue"), uno.createUnoStruct("com.sun.star.beans.NamedValue"))
            aPosSizeParm = self.aEnv.fromkeys(self.aEnv['POSSIZEPARM'])
            for i in range(1,4):
                if aPolars[i][0] > 0:
                    # print "straight line - ligne"
                    self.aEnv['UNOTYPE'] = "LineShape"
                    self.aEnv['ENTITYPE'] = "LINE"
                    self.aEnv['STEP'] = 1
                    self.aEnv['MODE'] = (self.aEnv['MODE'][0],) + ("VERTEX",)
                    self.aEnv['SHAPE'] = None
                    if i == 1:
                        res = self.xCompute.getInterLines2( aCoords[i-1], aCoords[i], aCoords[i], aCoords[i+1], 0, self.nStridePos)
                        aCoord[0]['X'] = res[2]
                        aCoord[0]['Y'] = res[3]
                    else:
                        aCoord[0]['X'] = aCoords[i]['X'] + long(self.nStridePos * math.sin(aPolars[i][1]))
                        aCoord[0]['Y'] = aCoords[i]['Y'] + long(self.nStridePos * math.cos(aPolars[i][1]))
                    if i == 3:
                        res = self.xCompute.getInterLines2( aCoords[i], aCoords[i+1], aCoords[i+1], aCoords[i+2], self.nStridePos, 0)
                        aCoord[1]['X'] = res[2]
                        aCoord[1]['Y'] = res[3]
                    else:
                        aCoord[1]['X'] = aCoords[i+1]['X'] + long(self.nStridePos * math.sin(aPolars[i][1]))
                        aCoord[1]['Y'] = aCoords[i+1]['Y'] + long(self.nStridePos * math.cos(aPolars[i][1]))
                    self.aEnv['COORDS'] = [aCoord[:2],]
                    self.aEnv['ORIGINCOORD'] = aCoord[0]
                    self.aEnv['SHAPESTATUS'] = "new"
                    for aKey in aPosSizeParm.iterkeys():
                        aPosSizeParm[aKey] = self.aEnv[aKey]
                    self.aEnv.update(self.xEntityPosSize.execute(aPosSizeParm))
                    self.aEnv['SHAPE'].setPropertyValue("LineStyle", uno.Enum("com.sun.star.drawing.LineStyle", "DASH"))
                    self.aEnv['SHAPE'].LineDash = aLineDash
                if i < 3:
                    # print " arc"
                    self.aEnv['CIRCLEKIND'] = ARC
                    self.aEnv['UNOTYPE'] = "EllipseShape"
                    self.aEnv['ENTITYPE'] = "ARC"
                    self.aEnv['STEP'] = 3
                    self.aEnv['MODE'] = (self.aEnv['MODE'][0],) + ("RADIUS", "STARTANGLE", "ENDANGLE")
                    self.aEnv['SHAPE'] = None
                    aCoord[0]['X'] = aCoords[i+1]['X']
                    aCoord[0]['Y'] = aCoords[i+1]['Y']
                    if self.nStridePos > 0:
                        aCoord[2]['X'] = aCoord[0]['X'] + abs(self.nStridePos) * math.cos(aPolars[i][1] - fStridePos)
                        aCoord[2]['Y'] = aCoord[0]['Y'] - abs(self.nStridePos) * math.sin(aPolars[i][1] - fStridePos)
                        aCoord[3]['X'] = aCoord[0]['X'] + abs(self.nStridePos) * math.cos(aPolars[i+1][1] - fStridePos)
                        aCoord[3]['Y'] = aCoord[0]['Y'] - abs(self.nStridePos) * math.sin(aPolars[i+1][1] - fStridePos)
                    else:
                        aCoord[3]['X'] = aCoord[0]['X'] + abs(self.nStridePos) * math.cos(aPolars[i][1] - fStridePos)
                        aCoord[3]['Y'] = aCoord[0]['Y'] - abs(self.nStridePos) * math.sin(aPolars[i][1] - fStridePos)
                        aCoord[2]['X'] = aCoord[0]['X'] + abs(self.nStridePos) * math.cos(aPolars[i+1][1] - fStridePos)
                        aCoord[2]['Y'] = aCoord[0]['Y'] - abs(self.nStridePos) * math.sin(aPolars[i+1][1] - fStridePos)
                    aCoord[1]['X'] = aCoord[2]['X']
                    aCoord[1]['Y'] = aCoord[2]['Y']
                    self.aEnv['COORDS'] = [aCoord,]
                    self.aEnv['ORIGINCOORD'] = aCoord[0]
                    self.aEnv['SHAPESTATUS'] = "new"
                    for aKey in aPosSizeParm.iterkeys():
                        aPosSizeParm[aKey] = self.aEnv[aKey]
                    self.aEnv.update(self.xEntityPosSize.execute(aPosSizeParm))
                    self.aEnv['SHAPE'].setPropertyValue("LineStyle", uno.Enum("com.sun.star.drawing.LineStyle", "DASH"))
                    self.aEnv['SHAPE'].LineDash = aLineDash
            # -------------------------------------------------------------------------------
            # print "Calculating length of path line - longueur de la ligne de foulee"
            # Splitting path and collar line - decomposition de la ligne de foulee et de jour
            # -------------------------------------------------------------------------------
            nCollarLength = []
            nStrideLength = []
            nCollarLength.append(max((long(aPolars[1][0]) - nStepWidth), 0))
            nCollarLength.append(long(aPolars[1][0]))
            nStrideLength.append(nCollarLength[0])
            nStrideLength.append(nCollarLength[1])
            # ------------------------------------
            nStrideLength.append(nStrideLength[1] + long(abs((((aPolars[2][1] - aPolars[1][1]) * math.sin(fStridePos) + 2 * math.pi) % (2 * math.pi)) * self.nStridePos)))
            nStrideLength.append(nStrideLength[2] + long(aPolars[2][0] / 2))
            nStrideLength.append(nStrideLength[3] + long(aPolars[2][0] / 2))
            nStrideLength.append(nStrideLength[4] + long(abs((((aPolars[3][1] - aPolars[2][1]) * math.sin(fStridePos) + 2 * math.pi) % (2 * math.pi)) * self.nStridePos)))
            nCollarLength.append(nCollarLength[1] + long(aPolars[2][0] / 2))
            nCollarLength.append(nCollarLength[2] + long(aPolars[2][0] / 2))
            # -------------------------------------
            nStrideLength.append(nStrideLength[5] + min(long(aPolars[3][0]), nStepWidth))
            nStrideLength.append(nStrideLength[5] + long(aPolars[3][0]))
            nCollarLength.append(nCollarLength[3] + min(long(aPolars[3][0]), nStepWidth))
            nCollarLength.append(nCollarLength[3] + long(aPolars[3][0]))
            # print "Calculating the tread length - calcul du giron"
            nTreadLength = long(float(nStrideLength[-1]) / (self.nStepsCount - 1))
            # print "Correcting regular length value - correction des longueurs de marches normales"
            if self.n1stStepBal > 0:
                nStrideLength[0] = (self.n1stStepBal - 2) * nTreadLength
                nCollarLength[0] = (self.n1stStepBal - 2) * nTreadLength
            else:
                nCollarLength[0] = (nCollarLength[0] / nTreadLength) * nTreadLength
                nStrideLength[0] = (nStrideLength[0] / nTreadLength) * nTreadLength
            if self.nLastStepBal > 0:
                nStrideLength[6] = nStrideLength[7] - (self.nStepsCount - self.nLastStepBal -1) * nTreadLength
                nCollarLength[4] = nCollarLength[5] - (self.nStepsCount - self.nLastStepBal - 1) * nTreadLength
            else:
                nStrideLength[6] = nStrideLength[7] - ((nStrideLength[7] - nStrideLength[6]) / nTreadLength) * nTreadLength
                nCollarLength[4] = nCollarLength[5] - ((nCollarLength[5] - nCollarLength[4]) / nTreadLength) * nTreadLength
            #print "\n".join(["collet > %d ; foulee > %d" % (c, f) for c,f in zip(nCollarLength+[0,0], nStrideLength)])
            ## -------------------------------------
            # print " Drawing de stairs - trace des marches"
            ## -------------------------------------
            aPoint = uno.createUnoStruct("com.sun.star.awt.Point")
            aSize = uno.createUnoStruct("com.sun.star.awt.Size")
            aHomogenMatrix = uno.createUnoStruct("com.sun.star.drawing.HomogenMatrix3")
            aHomogenMatrix.Line1 = uno.createUnoStruct("com.sun.star.drawing.HomogenMatrixLine3")
            aHomogenMatrix.Line2 = uno.createUnoStruct("com.sun.star.drawing.HomogenMatrixLine3")
            aHomogenMatrix.Line3 = uno.createUnoStruct("com.sun.star.drawing.HomogenMatrixLine3")
            # print "nombre de marches %d" % self.nStepsCount
            for i in range(self.nStepsCount - 2):
                nStrideDist = (i + 1) * nTreadLength
                ## --------------------------------
                if nStrideDist <= nStrideLength[0] or nStrideDist >= nStrideLength[6]:
                    #print "Case 1 and 8 - Cas 1 et 8 ####"
                    n = -(nStrideDist <= nStrideLength[0]) + (nStrideDist >= nStrideLength[6])
                    nDistance = nStrideDist - nStrideLength[5] * (n > 0)
                    nIndex = n + 2
                    fAngle = aPolars[nIndex][1]
                    # print "Coordinates on collar line - coordonnees sur la ligne de jour (collet)"
                    aCoord[0]['X'] = aCoords[nIndex]['X'] + long(nDistance * math.cos(fAngle))
                    aCoord[0]['Y'] = aCoords[nIndex]['Y'] - long(nDistance * math.sin(fAngle))
                    # print "Coordinates on stride line - coordonnees sur la ligne de foulee"
                    ## no need but text position - pas besoin sauf pour le texte
                    aCoord[1]['X'] = aCoords[nIndex]['X'] + long(nDistance * math.cos(fAngle) + self.nStridePos * math.sin(fAngle))
                    aCoord[1]['Y'] = aCoords[nIndex]['Y'] - long(nDistance * math.sin(fAngle) - self.nStridePos * math.cos(fAngle))
                    # print "Coordinates on outside line - coordonnees sur la ligne d'echiffre"
                    aCoord[2]['X'] = aCoord[0]['X'] + long(nStepWidth * math.sin(fStridePos) * math.sin(fAngle))
                    aCoord[2]['Y'] = aCoord[0]['Y'] + long(nStepWidth * math.sin(fStridePos) * math.cos(fAngle))
                    fAngle = self.xCompute.toPolarCoord(aCoord[0], aCoord[1])[1]
                ## --------------------------------
                elif nStrideDist <= nStrideLength[1] or nStrideDist >= nStrideLength[5]:
                    #print "Case 2 and 7 - Cas 2 et 7 ####"
                    n = -(nStrideDist <= nStrideLength[1]) + (nStrideDist >= nStrideLength[5])
                    nDistance = nStrideDist - nStrideLength[5] * (n > 0)
                    nIndex = n + 2
                    fAngle = aPolars[nIndex][1]
                    # print "Coordinates on stride line - coordonnees sur la ligne de foulee"
                    aCoord[1]['X'] = aCoords[nIndex]['X'] + long(nDistance * math.cos(fAngle) + self.nStridePos * math.sin(fAngle))
                    aCoord[1]['Y'] = aCoords[nIndex]['Y'] - long(nDistance * math.sin(fAngle) - self.nStridePos * math.cos(fAngle))
                    # print "Coordinates on collar line - coordonnees sur la ligne de jour (collet)"
                    nDistance = nCollarLength[0]*(n < 0) + (nCollarLength[4] - nCollarLength[3]) * (n > 0)
                    nDistance = abs(nDistance - n * self.xCompute.getCollarX( abs(nCollarLength[2] - nCollarLength[2+2*n]), abs(nStrideLength[3] - nStrideLength[3+3*n]), abs(nStrideDist - nStrideLength[3+3*n])))
                    aCoord[0]['X'] = aCoords[nIndex]['X'] + long(nDistance * math.cos(fAngle))
                    aCoord[0]['Y'] = aCoords[nIndex]['Y'] - long(nDistance * math.sin(fAngle))
                    # print "Coordinates on outside line - coordonnees sur la ligne d'echiffre"
                    fAngle = self.xCompute.toPolarCoord(aCoord[0], aCoord[1])[1]
                    nIndex = 6 * (n > 0)
                    fOutsideAngle = (aPolars[nIndex - 1][1] + math.pi) % (2 * math.pi)
                    rep = self.xCompute.getInterLines(aCoord[0], fAngle, aCoords[nIndex], fOutsideAngle)
                    if rep is None:
                        self.xMessage.execute(self.aTranslation[6], "errorbox")
                        self.aEnv['CONTOURLIST'] = None
                        self.aEnv['STAGE'] = 0
                        return self.aEnv
                    else:
                        nDistance = rep[0]
                    aCoord[2]['X'] = aCoord[0]['X'] + long(nDistance * math.cos(fAngle))
                    aCoord[2]['Y'] = aCoord[0]['Y'] - long(nDistance * math.sin(fAngle))
                ## --------------------------------
                elif nStrideDist <= nStrideLength[2] or nStrideDist >= nStrideLength[4]:
                    #print "Case 3 and 6 - Cas 3 et 6 ####"
                    n = -(nStrideDist <= nStrideLength[2]) + (nStrideDist >= nStrideLength[4])
                    nIndex = (n < 0) + 3 * (n > 0)
                    nDistance = nStrideDist - nStrideLength[1] * (n < 0) - nStrideLength[5] * (n > 0)
                    fAngle = aPolars[nIndex][1] - fStridePos + float(nDistance) / self.nStridePos
                    # print "Coordinates on stride line - coordonnees sur la ligne de foulee"
                    nIndex = 2 * (n < 0) + 3 * (n > 0)
                    aCoord[1]['X'] = aCoords[nIndex]['X'] + long(abs(self.nStridePos) * math.cos(fAngle))
                    aCoord[1]['Y'] = aCoords[nIndex]['Y'] - long(abs(self.nStridePos) * math.sin(fAngle))
                    # print "Coordinates on collar line - coordonnees sur la ligne de jour (collet)"
                    nDistance = nCollarLength[0]*(nStrideDist <= nStrideLength[2]) + (nCollarLength[4] - nCollarLength[3]) * (nStrideDist >= nStrideLength[4])
                    nDistance = nDistance - n * self.xCompute.getCollarX(abs(nCollarLength[2] - nCollarLength[2+2*n]), abs(nStrideLength[3] - nStrideLength[3+3*n]), abs(nStrideDist - nStrideLength[3+3*n]))
                    if nDistance <= nCollarLength[1] and nDistance >= 0:
                        nIndex = (nStrideDist <= nStrideLength[2]) + 3 * (nStrideDist >= nStrideLength[4])
                        fAngle = aPolars[nIndex][1]
                        aCoord[0]['X'] = aCoords[nIndex]['X'] + long(nDistance * math.cos(fAngle))
                        aCoord[0]['Y'] = aCoords[nIndex]['Y'] - long(nDistance * math.sin(fAngle))
                    else:
                        nDistance = nDistance - nCollarLength[1] * (nStrideDist <= nStrideLength[2]) + aPolars[2][0] * (nStrideDist >= nStrideLength[4])
                        fAngle = aPolars[2][1]
                        aCoord[0]['X'] = aCoords[2]['X'] + long(nDistance * math.cos(fAngle))
                        aCoord[0]['Y'] = aCoords[2]['Y'] - long(nDistance * math.sin(fAngle))
                    # print "Coordinates on outside line - coordonnees sur la ligne d'echiffre"
                    fAngle = self.xCompute.toPolarCoord(aCoord[0], aCoord[1])[1]
                    nIndex = 5 * (n > 0)
                    fOutsideAngle = ((aPolars[nIndex - 1][1] + math.pi) % (2 * math.pi)) * (n < 0) + (aPolars[nIndex][1]) * (n > 0)
                    rep = self.xCompute.getInterLines(aCoord[0], fAngle, aCoords[nIndex], fOutsideAngle)
                    if rep is None:
                        nDistance = aPolars[nIndex-(n<0)][0] + abs(self.nStridePos)
                    else:
                        nDistance = rep[1]
                    # --------------------------------------
                    if nDistance <= aPolars[nIndex-(n<0)][0]:
                        aCoord[2]['X'] = aCoords[nIndex]['X'] + long(nDistance * math.cos(fOutsideAngle))
                        aCoord[2]['Y'] = aCoords[nIndex]['Y'] - long(nDistance * math.sin(fOutsideAngle))
                    else:
                        fOutsideAngle = (aPolars[6][1] + math.pi) % (2 * math.pi)
                        rep = self.xCompute.getInterLines(aCoord[0], fAngle, aCoords[7], fOutsideAngle)
                        if rep is None:
                            self.xMessage.execute(self.aTranslation[6], "errorbox")
                            self.aEnv['CONTOURLIST'] = None
                            self.aEnv['STAGE'] = 0
                            return self.aEnv
                        else:
                            nDistance = rep[0]
                        aCoord[2]['X'] = aCoord[0]['X'] + long(nDistance * math.cos(fAngle))
                        aCoord[2]['Y'] = aCoord[0]['Y'] - long(nDistance * math.sin(fAngle))
                ## --------------------------------
                elif nStrideDist <= nStrideLength[3] or nStrideDist > nStrideLength[3]:
                    #print "Case 4 and 5 - Cas 4 et 5 ####"
                    nDistance = nStrideDist - nStrideLength[2]
                    # print "Coordinates on stride line - coordonnees sur la ligne de foulee"
                    fAngle = aPolars[2][1]
                    aCoord[1]['X'] = aCoords[2]['X'] + long(nDistance * math.cos(fAngle) + self.nStridePos * math.sin(fAngle))
                    aCoord[1]['Y'] = aCoords[2]['Y'] - long(nDistance * math.sin(fAngle) - self.nStridePos * math.cos(fAngle))
                    # print "Coordinates on collar line - coordonnees sur la ligne de jour (collet)"
                    n = -(nStrideDist <= nStrideLength[3]) + (nStrideDist > nStrideLength[3])
                    nDistance = (nCollarLength[1] - nCollarLength[0]) * (n < 0) + (nCollarLength[4] - nCollarLength[3]) * (n > 0)
                    nDistance = self.xCompute.getCollarX(abs(nCollarLength[2] - nCollarLength[2+2*n]), abs(nStrideLength[3] - nStrideLength[3+3*n]), abs(nStrideDist - nStrideLength[3+3*n])) - nDistance
                    nDistance = -n * nDistance + aPolars[2][0] * (n > 0)
                    aCoord[0]['X'] = aCoords[2]['X'] + long(nDistance * math.cos(fAngle))
                    aCoord[0]['Y'] = aCoords[2]['Y'] - long(nDistance * math.sin(fAngle))
                    # print "Coordinates on outside line - coordonnees sur la ligne d'echiffre"
                    fAngle = self.xCompute.toPolarCoord(aCoord[0], aCoord[1])[1]
                    fOutsideAngle = (aPolars[6][1] + math.pi) % (2 * math.pi)
                    rep = self.xCompute.getInterLines(aCoord[0], fAngle, aCoords[7], fOutsideAngle)
                    if rep is None:
                        self.xMessage.execute(self.aTranslation[6], "errorbox")
                        self.aEnv['CONTOURLIST'] = None
                        self.aEnv['STAGE'] = 0
                        return self.aEnv
                    else:
                        nDistance = rep[0]
                    aCoord[2]['X'] = aCoord[0]['X'] + long(nDistance * math.cos(fAngle))
                    aCoord[2]['Y'] = aCoord[0]['Y'] - long(nDistance * math.sin(fAngle))
                # -----------------------------------------
                self.aEnv['UNOTYPE'] = "LineShape"
                self.aEnv['ENTITYPE'] = "LINE"
                self.aEnv['STEP'] = 1
                self.aEnv['MODE'] = (self.aEnv['MODE'][0],) + ("VERTEX",)
                self.aEnv['SHAPE'] = None
                self.aEnv['COORDS'] = [[aCoord[0], aCoord[2]]]
                self.aEnv['ORIGINCOORD'] = aCoord[0]
                aPosSizeParm = self.aEnv.fromkeys(self.aEnv['POSSIZEPARM'])
                for aKey in aPosSizeParm.iterkeys():
                    aPosSizeParm[aKey] = self.aEnv[aKey]
                self.aEnv.update(self.xEntityPosSize.execute(aPosSizeParm))
                ##
                # print "Step number drawing - affichage du numero de marche"
                ##
                aTextShape = self.aEnv['CONTROLLER'].Model.createInstance("com.sun.star.drawing.TextShape")
                # print "create text shape"
                aPoint.X, aPoint.Y = aCoord[1]['X'], aCoord[1]['Y']
                aTextShape.setPosition(aPoint)
                aSize.Width = 200 * long(math.log10(i+2) + 1)
                aSize.Height = 350
                aTextShape.setSize(aSize)
                # print "set position and size"
                # aTextShape.setPropertyValue("TextFitToSize", uno.Enum("com.sun.star.drawing.TextFitToSizeType", "PROPORTIONAL"))
                if fStridePos < 0:
                    fAngle = (fAngle + math.pi) % (2 * math.pi)
                # print "add to group shape"
                #print aTextShape.Transformation
                #aHomogenMatrix.Line1.Column1 = math.cos(fAngle)
                #aHomogenMatrix.Line1.Column2 = -math.sin(fAngle)
                #aHomogenMatrix.Line1.Column3 = 0.0
                #aHomogenMatrix.Line2.Column1 = math.sin(fAngle)
                #aHomogenMatrix.Line2.Column2 = math.cos(fAngle)
                #aHomogenMatrix.Line2.Column3 = 0.0
                #aHomogenMatrix.Line3.Column1 = 0.0
                #aHomogenMatrix.Line3.Column2 = 0.0
                #aHomogenMatrix.Line3.Column3 = 1.0
                #aTextShape.Transformation = aHomogenMatrix
                aTextShape.RotateAngle = long(fAngle * 18000 / math.pi)
                # print "text rotation"
                self.aEnv['GROUPSHAPE'].add(aTextShape)
                aTextShape.TextAutoGrowWidth = True
                aTextCursor = aTextShape.createTextCursor()
                aTextShape.insertString(aTextCursor, "%d" % (i+2), False)
                aTextShape.CharHeight = self.aEnv['CHARHEIGHT']
                # print "insert string in text shape"
            self.aEnv['SHAPE'] = None
            self.aEnv['COORDS'] = None
            ##
            # print "First step number - numero de la premiere marche"
            ##
            aTextShape = self.aEnv['MODEL'].createInstance("com.sun.star.drawing.TextShape")
            ## aCoord[1]['X'], aCoord[1]['Y'] = [long((aCoords[0]['X'] + aCoords[1]['X'])/2), long((aCoords[0]['Y'] + aCoords[1]['Y'])/2)]
            aCoord[1]['X'] = aCoords[1]['X'] - long(abs(self.nStridePos) * math.cos(aPolars[0][1]))
            aCoord[1]['Y'] = aCoords[1]['Y'] + long(abs(self.nStridePos) * math.sin(aPolars[0][1]))
            aPoint.X, aPoint.Y = aCoord[1]['X'], aCoord[1]['Y']
            aTextShape.setPosition(aPoint)
            aSize.Width = 200
            aSize.Height = 250
            aTextShape.setSize(aSize)
            # aTextShape.setPropertyValue("TextFitToSize", uno.Enum("com.sun.star.drawing.TextFitToSizeType", "PROPORTIONAL"))
            fAngle = aPolars[0][1]
            if fStridePos > 0:
                fAngle = (fAngle + math.pi) % (2 * math.pi)
            aTextShape.RotateAngle = long(fAngle * 18000 / math.pi)
            self.aEnv['GROUPSHAPE'].add(aTextShape)
            aTextShape.TextAutoGrowWidth = True
            aTextCursor = aTextShape.createTextCursor()
            aTextShape.insertString(aTextCursor, "1", False)
            aTextShape.CharHeight = self.aEnv['CHARHEIGHT']
            ##
            # print "Last step number - numero de la derniere marche"
            ##
            aTextShape = self.aEnv['MODEL'].createInstance("com.sun.star.drawing.TextShape")
            ## aCoord[1]['X'], aCoord[1]['Y'] = [long((aCoords[4]['X'] + aCoords[5]['X'])/2), long((aCoords[4]['Y'] + aCoords[5]['Y'])/2)]
            aCoord[1]['X'] = aCoords[4]['X'] + long(abs(self.nStridePos) * math.cos(aPolars[4][1]))
            aCoord[1]['Y'] = aCoords[4]['Y'] - long(abs(self.nStridePos) * math.sin(aPolars[4][1]))
            aPoint.X, aPoint.Y = aCoord[1]['X'], aCoord[1]['Y']
            aTextShape.setPosition(aPoint)
            aSize.Width = 400
            aSize.Height = 250
            aTextShape.setSize(aSize)
            # aTextShape.setPropertyValue("TextFitToSize", uno.Enum("com.sun.star.drawing.TextFitToSizeType", "PROPORTIONAL"))
            fAngle = aPolars[4][1]
            if fStridePos < 0:
                fAngle = (fAngle + math.pi) % (2 * math.pi)
            aTextShape.RotateAngle = long(fAngle * 18000 / math.pi)
            self.aEnv['GROUPSHAPE'].add(aTextShape)
            aTextShape.TextAutoGrowWidth = True
            aTextCursor = aTextShape.createTextCursor()
            aTextShape.insertString(aTextCursor, "%d" % self.nStepsCount, False)
            aTextShape.CharHeight = self.aEnv['CHARHEIGHT']
            ##
            # print "Stride dimension - cotation du giron"
            ##
            aDimShape = self.aEnv['MODEL'].createInstance("com.sun.star.drawing.MeasureShape")
            aCoord[0]['X'], aCoord[0]['Y'] = [long((aCoords[0]['X'] + aCoords[1]['X'])/2), long((aCoords[0]['Y'] + aCoords[1]['Y'])/2)]
            aCoord[1]['X'] = aCoord[0]['X'] + long(nTreadLength * math.cos(aPolars[1][1]))
            aCoord[1]['Y'] = aCoord[0]['Y'] - long(nTreadLength * math.sin(aPolars[1][1]))
            self.aEnv['GROUPSHAPE'].add(aDimShape)
            aPoint.X, aPoint.Y = aCoord[0]['X'], aCoord[0]['Y']
            aDimShape.StartPosition = aPoint
            aPoint.X, aPoint.Y = aCoord[1]['X'], aCoord[1]['Y']
            aDimShape.EndPosition = aPoint
            aDimShape.CharHeight = 8.0
            aDimShape.MeasureShowUnit = False
            self.aEnv['STOPJOB'] = True
            return self.aEnv
    # *****************************************************
    # remove temporary shapes
    # *****************************************************
    def removeTempShapes(self):
        if not self.aEnv['SPOTTEDID'] is None:
            aGroupShape = self.aEnv['CONTROLLER'].getCurrentPage()
            if len(self.aEnv['SPOTTEDID']) > 2:
                for i in self.aEnv['SPOTTEDID'][:-2]:
                    aGroupShape = aGroupShape.getByIndex(i)
            aGroupShape.remove(self.aEnv['SHAPE'])
            self.aEnv['SHAPE'] = None
            self.aEnv['STEP'] = 0
            self.aEnv['COORDS'] = None
            self.aEnv['SPOTTEDID'] = None
        return
    # ******************************************
    # set entity env and get coordinates
    # ******************************************
    def setEntityEnvAndCoord(self, cadoooShape):
        aCoords = [[]]
        # ...........................................
        if cadoooShape[1] in ('LineShape',):
            self.aEnv['UNOTYPE'] = "LineShape"
            self.aEnv['ENTITYPE'] = "LINE"
            self.aEnv['STEP'] = 1
            self.aEnv['MODE'] = (self.aEnv['MODE'][0],) + ("VERTEX",)
            aCoords[0].extend([{'X': 0, 'Y': 0} for i in range(2)])
            aCoords[0][0]['X'], aCoords[0][0]['Y'] = cadoooShape[2]['X'], cadoooShape[2]['Y']
            aCoords[0][1]['X'], aCoords[0][1]['Y'] = cadoooShape[3]['X'], cadoooShape[3]['Y']
        # ........................................
        elif cadoooShape[1] in ('CircleShape',):
            self.aEnv['CIRCLEKIND'] = FULL
            self.aEnv['UNOTYPE'] = "EllipseShape"
            self.aEnv['ENTITYPE'] = "CIRCLE"
            self.aEnv['STEP'] = 1
            self.aEnv['MODE'] = (self.aEnv['MODE'][0],) + ("RADIUS",)
            aCoords[0].extend([{'X': 0, 'Y': 0} for i in range(2)])
            aCoords[0][0]['X'] = cadoooShape[2]['X']
            aCoords[0][0]['Y'] = cadoooShape[2]['Y']
            aPoint = dict(self.xCompute("PointOnEllipse", None, math.pi/4, *cadoooShape[2:5])[0])
            aCoords[0][1]['X'] = aPoint['X']
            aCoords[0][1]['Y'] = aPoint['Y']
        # .........................................................
        elif cadoooShape[1] in ("ArcCircleShape",):
            self.aEnv['CIRCLEKIND'] = ARC
            self.aEnv['UNOTYPE'] = "EllipseShape"
            self.aEnv['ENTITYPE'] = "ARC"
            self.aEnv['STEP'] = 3
            self.aEnv['MODE'] = (self.aEnv['MODE'][0],) + ("RADIUS", "STARTANGLE", "ENDANGLE")
            aCoords[0].extend([{'X': 0, 'Y': 0} for i in range(4)])
            aCoords[0][0]['X'] = cadoooShape[2]['X']
            aCoords[0][0]['Y'] = cadoooShape[2]['Y']
            aPoint = dict(self.xCompute("PointOnEllipse", None, math.pi/4, *cadoooShape[2:5])[0])
            aCoords[0][1]['X'] = aPoint['X']
            aCoords[0][1]['Y'] = aPoint['Y']
            aPoint = dict(self.xCompute("PointOnEllipse", None, cadoooShape[5], *cadoooShape[2:5])[0])
            aCoords[0][2]['X'] = aPoint['X']
            aCoords[0][2]['Y'] = aPoint['Y']
            aPoint = dict(self.xCompute("PointOnEllipse", None, cadoooShape[6], *cadoooShape[2:5])[0])
            aCoords[0][3]['X'] = aPoint['X']
            aCoords[0][3]['Y'] = aPoint['Y']
        # .........................................................
        elif cadoooShape[1] in ("ArcEllipseShape",):
            self.aEnv['CIRCLEKIND'] = ARC
            self.aEnv['UNOTYPE'] = "EllipseShape"
            self.aEnv['ENTITYPE'] = "ELLARC"
            self.aEnv['STEP'] = 4
            self.aEnv['MODE'] = (self.aEnv['MODE'][0],) + ("RADIUS", "RADIUS2", "STARTANGLE", "ENDANGLE")
            aCoords[0].extend([{'X': 0, 'Y': 0} for i in range(5)])
            aCoords[0][0]['X'] = cadoooShape[2]['X']
            aCoords[0][0]['Y'] = cadoooShape[2]['Y']
            aPoint = dict(self.xCompute("PointOnEllipse", None, 0.0, *cadoooShape[2:5])[0])
            aCoords[0][1]['X'] = aPoint['X']
            aCoords[0][1]['Y'] = aPoint['Y']
            aPoint = dict(self.xCompute("PointOnEllipse", None, math.pi/2, *cadoooShape[2:5])[0])
            aCoords[0][2]['X'] = aPoint['X']
            aCoords[0][2]['Y'] = aPoint['Y']
            aPoint = dict(self.xCompute("PointOnEllipse", None, cadoooShape[5], *cadoooShape[2:5])[0])
            aCoords[0][3]['X'] = aPoint['X']
            aCoords[0][3]['Y'] = aPoint['Y']
            aPoint = dict(self.xCompute("PointOnEllipse", None, cadoooShape[6], *cadoooShape[2:5])[0])
            aCoords[0][4]['X'] = aPoint['X']
            aCoords[0][4]['Y'] = aPoint['Y']
        else:
            aCoords = None
        return aCoords
