#
# -*- coding: latin-1 -*-
## ********************************************************************************
## ucskeyhandler dimanche 28 juin 2009, 16:01:47 (UTC+0200)

## Copyright (C) 2008-2009 Grard Deneux
## gerard.deneux@free.fr

## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.

## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
## Lesser General Public License for more details.
## http://www.opensource.org/licenses/lgpl-license.php

## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
## ******************************************************************************
try:
    import math
    from com.sun.star.awt.KeyModifier import SHIFT
except ImportError:
    print "probleme d'import"
#===============================
# gestion des evenements clavier
# handling keyboard events
#===============================
class UcsKeyHandler():
    """gestion des evenements clavier
    handling keyboard events"""
    def __init__(self, ctx):
        from fr.free.deneux.gerard.cadooo.datainputbox import DataInputBox
        from fr.free.deneux.gerard.cadooo.entityposize import EntityPosSize
        from fr.free.deneux.gerard.cadooo.compute import Compute
        from fr.free.deneux.gerard.cadooo.msgl10n import MsgL10n
        self.ctx = ctx
        # Datum needed by coordinates input form
        self.aCoordData = {}
        self.aCoordData["TITLE"]  = "No Title"
        self.aCoordData["FLAG"] = 0
        self.aCoordData["COMMENT"] = "No comment"
        self.aCoordData["X"] = 0
        self.aCoordData["XUNIT"] = "m"
        self.aCoordData["Y"] = 0
        self.aCoordData["YUNIT"] = "m"
        self.aCoordData["RETURN"] = False
        self.xCoordInput = DataInputBox( self.ctx)
        self.xEntityPosSize = EntityPosSize(self.ctx)
        self.xCompute = Compute(self.ctx)
        self.xMsgL10n = MsgL10n(self.ctx)
    # ==================================
    # actions clavier - keyboard actions
    # aNamedValues as tuple of NamedValue
    #   Name         Value
    # KEYCODE     integer
    # L2OSTATE    tuple of 5 integers (0 or 1)
    # DRAWSCALE
    # COORDS
    # STEP
    # VALIDORIGIN
    # DRAWSCALE
    # ORIGINCOORD
    # BORDERCOORD
    # CANCELJOB
    # ORTHOMODE
    # SHAPE
    # ==================================
    def execute(self, aArgs):
        aEnv = {}
        aEnv.update(aArgs)
        lKey = aEnv['MAGNETKEY'].keys()
        lKey.sort()
        # ----------------------------------------------------------------------------------------------------
        if aEnv['KEYEVENT']['COD'] in lKey and not aEnv['KEYEVENT']['MOD']:
            # e,   i,   m,   p,   t => link 2 entities
            i = lKey.index(aEnv['KEYEVENT']['COD'])
            bKHState = list(aEnv['L2OSTATE'])
            bKHState[i] = not bKHState[i]
            aEnv['L2OSTATE'] = tuple(bKHState)
            aStatusBarCheckBox = aEnv['STATUSBARCONTROL'][aEnv['KEYEVENT']['COD']]
            if aStatusBarCheckBox.getState() != aEnv['L2OSTATE'][i]:
                aStatusBarCheckBox.setState(aEnv['L2OSTATE'][i])
        # -------------------------------------------------------------------------
        elif aEnv['KEYEVENT']['COD'] in ('AXOCTRL',) and not aEnv['KEYEVENT']['MOD']:
            #                                  a => axonometric mode
            aEnv['ORTHOMODE'] = not aEnv['ORTHOMODE']
            aStatusBarCheckBox = aEnv['STATUSBARCONTROL']['AXOCTRL']
            if aStatusBarCheckBox.getState() != aEnv['ORTHOMODE']:
                aStatusBarCheckBox.setState(aEnv['ORTHOMODE'])
        # -----------------------------------------------------------------------------
        elif aEnv['KEYEVENT']['COD'] in ('CAR', 'POL') and not aEnv['KEYEVENT']['MOD']:
            # ..............................c, ..k
            if aEnv['COORDS'] is None:
                aEnv['COORDS'] = [[{'X': 0, 'Y': 0}]]
            # ------------------------------------
            if aEnv['KEYEVENT']['COD'] in ('CAR',):
                #                             c => rectangular coordinates requested
                aMouseParm = aEnv.fromkeys(aEnv['MOUSEPARM'])
                for aKey in aMouseParm.iterkeys():
                    aMouseParm[aKey] = aEnv[aKey]
                aMouseCoord = self.xCompute.transformMouseCoord( aMouseParm)
                aMouseCoord = self.xCompute.transform2UCS( aMouseCoord, aEnv['UCS'])
                sTextList = ("k2", "k3")
                self.sText = self.xMsgL10n.getText(sTextList)
                self.aCoordData["TITLE"]  = self.sText[0]
                self.aCoordData["FLAG"] = 1
                self.aCoordData["COMMENT"] = self.sText[1]
                self.aCoordData["RETURN"] = False
                self.aCoordData["X"] = self.xCompute.transform2UserUnit( aMouseCoord['X'], aEnv['UNITS'], aEnv['UNITSRATIO'], aEnv['MEASUREUNIT'], aEnv['DRAWSCALE'])
                self.aCoordData["Y"] = self.xCompute.transform2UserUnit( aMouseCoord['Y'], aEnv['UNITS'], aEnv['UNITSRATIO'], aEnv['MEASUREUNIT'], aEnv['DRAWSCALE'])
                aInputData = aEnv.fromkeys(aEnv['INPUTDATAPARM'])
                for aKey in aInputData.iterkeys():
                    aInputData[aKey] = aEnv[aKey]
                self.aCoordData.update(aInputData)
                self.aCoordData.update(self.xCoordInput.execute(self.aCoordData))
                # ---------------------------
                if self.aCoordData["RETURN"]:
                    # -------------------------------
                    if self.aCoordData["FLAG"] & 0x2: # Delta coordinates ?
                        aEnv['COORDS'][-1][-1]['X'] = self.aCoordData['X'] + aEnv['UCS'][0][2]
                        aEnv['COORDS'][-1][-1]['Y'] = self.aCoordData['Y'] + aEnv['UCS'][1][2]
                    else:
                        aEnv['COORDS'][-1][-1]['X'] = self.aCoordData['X'] + aEnv['BORDERCOORD'][0]
                        aEnv['COORDS'][-1][-1]['Y'] = self.aCoordData['Y'] + aEnv['BORDERCOORD'][1]
                    if aEnv['STEP'] == 0:
                        aEnv['COORDS'][0].append({'X': 0, 'Y': 0})
                    if aEnv['STEP'] == 1:
                        aEnv['COORDS'].append(aEnv['COORDS'][0])
            # -----------------------------------------
            elif aEnv['KEYEVENT']['COD'] in ('POL',):
                #                               k => polar coordinates requested
                aMouseParm = aEnv.fromkeys(aEnv['MOUSEPARM'])
                for aKey in aMouseParm.iterkeys():
                    aMouseParm[aKey] = aEnv[aKey]
                aMouseCoord = self.xCompute.transformMouseCoord(aMouseParm)
                aPolarCoord = self.xCompute.toPolarCoord( {'X': aEnv['UCS'][0][2], 'Y': aEnv['UCS'][1][2]}, aMouseCoord)
                sTextList = ("k2", "k4")
                self.sText = self.xMsgL10n.execute(sTextList)
                self.aCoordData["TITLE"]  = self.sText[0]
                self.aCoordData["FLAG"] = 5
                self.aCoordData["COMMENT"] = self.sText[1]
                self.aCoordData["RETURN"] = False
                self.aCoordData["X"] = self.xCompute.transform2UserUnit( aPolarCoord[0], aEnv['UNITS'], aEnv['UNITSRATIO'], aEnv['MEASUREUNIT'], aEnv['DRAWSCALE'])
                self.aCoordData["Y"] = "%7.2f" % math.degrees((aPolarCoord[1] - math.atan2(aEnv['UCS'][0][1], aEnv['UCS'][0][0])) % 2*math.pi)
                aInputData = aEnv.fromkeys(aEnv['INPUTDATAPARM'])
                for aKey in aInputData.iterkeys():
                    aInputData[aKey] = aEnv[aKey]
                self.aCoordData.update(aInputData)
                self.aCoordData.update(self.aCoordInput.execute(self.aCoordData))
                # ---------------------------
                if self.aCoordData["RETURN"]:
                    aEnv['COORDS'][-1][-1]['X'] = long(round(self.aCoordData['X'] * math.cos(self.aCoordData["Y"])))
                    aEnv['COORDS'][-1][-1]['Y'] = long(round(self.aCoordData['X'] * math.sin(-self.aCoordData["Y"])))
                    # -------------------------------
                    if self.aCoordData["FLAG"] & 0x2: # delta coordinates ?
                        aEnv['COORDS'][-1][-1]['X'] = self.aCoordData['X'] + aEnv['UCS'][0][2]
                        aEnv['COORDS'][-1][-1]['Y'] = self.aCoordData['Y'] + aEnv['UCS'][1][2]
                    else:
                        aEnv['COORDS'][-1][-1]['X'] = aEnv['COORDS'][-1][-1]['X'] + aEnv['BORDERCOORD'][0]
                        aEnv['COORDS'][-1][-1]['Y'] = aEnv['COORDS'][-1][-1]['Y'] + aEnv['BORDERCOORD'][1]
                    if aEnv['STEP'] == 0:
                        aEnv['COORDS'][0].append({'X': 0, 'Y': 0})
                    if aEnv['STEP'] == 1:
                        aEnv['COORDS'].append(aEnv['COORDS'][0])
            # ---------------------
            if aEnv['STEP'] != 0:
                aPosSizeParm = aEnv.fromkeys(aEnv['POSSIZEPARM'])
                for aKey in aPosSizeParm.iterkeys():
                    aPosSizeParm[aKey] = aEnv[aKey]
                aEnv.update(self.xEntityPosSize.execute(aPosSizeParm))
            aEnv['ORIGINCOORD'].update(aEnv['COORDS'][0][0])
            aEnv['VALIDORIGIN'] = True
            aEnv['STEP'] = aEnv['STEP'] % aEnv['STEPS'] + 1
            # -------------------------------------
            if aEnv['STEP'] == aEnv['STEPS']:
                fAngleX = self.compute("toPolarCoord", aEnv['COORDS'][0][0], aEnv['COORDS'][0][1])[1]
                fAngleY = self.compute("toPolarCoord", aEnv['COORDS'][1][0], aEnv['COORDS'][1][1])[1]
                aEnv['UCS'] = ((math.cos(fAngleX), -math.cos(fAngleY), aShapeCoords[0][0]['X']), (math.sin(-fAngleX), math.sin(-fAngleY), aShapeCoords[0][0]['Y']), (0.0, 0.0, 1.0))
                aEnv['GROUPSHAPE'].remove(aEnv['SHAPE'])
                aEnv['SHAPE'] = aEnv['UCSSHAPE']
                aPosSizeParm = aEnv.fromkeys(aEnv['POSSIZEPARM'])
                for aKey in aPosSizeParm.iterkeys():
                    aPosSizeParm[aKey] = aEnv[aKey]
                aEnv.update(self.xEntityPosSize.execute(aPosSizeParm))
                aEnv['SHAPE'].setName("ucs")
                aEnv['SHAPE'].Style = aEnv['STANDARDSTYLE']
                aEnv['SHAPE'].LineColor = aEnv['STANDARDSTYLE'].LineColor
                aEnv['SHAPE'].LineWidth = aEnv['STANDARDSTYLE'].LineWidth
                aEnv['STOPJOB'] = True
        # -------------------------------------------------------------------------
        elif aEnv['KEYEVENT']['COD'] in ('DEF', 'DIA') and not aEnv['KEYEVENT']['MOD']:
            # ..............................d (set to default)
            aEnv['UCS'] = ((1.0, 0.0, aEnv['BORDERCOORD'][0]), (0.0, 1.0, aEnv['BORDERCOORD'][1]), (0.0, 0.0, 1.0))
            aEnv['COORDS'] = [[{'X': aEnv['BORDERCOORD'][0], 'Y': aEnv['BORDERCOORD'][1]}, {'X': aEnv['BORDERCOORD'][0] + 250, 'Y': aEnv['BORDERCOORD'][1]}],\
                       [{'X': aEnv['BORDERCOORD'][0], 'Y': aEnv['BORDERCOORD'][1]}, {'X': aEnv['BORDERCOORD'][0], 'Y': aEnv['BORDERCOORD'][1] + 250}]]
            aEnv['STEP'] = 2
            aEnv['GROUPSHAPE'].remove(aEnv['SHAPE'])
            aEnv['SHAPE'] = aEnv['UCSSHAPE']
            aPosSizeParm = aEnv.fromkeys(aEnv['POSSIZEPARM'])
            for aKey in aPosSizeParm.iterkeys():
                aPosSizeParm[aKey] = aEnv[aKey]
            aEnv.update(self.xEntityPosSize.execute(aPosSizeParm))
            aEnv['SHAPE'] = aEnv['STANDARDSTYLE']
            aEnv['SHAPE'].LineColor = aEnv['STANDARDSTYLE'].LineColor
            aEnv['SHAPE'].LineWidth = aEnv['STANDARDSTYLE'].LineWidth
            aEnv['SHAPE'].setName("ucs")
            aEnv['STOPJOB'] = True
        # -------------------------------------------------------------------------
        elif aEnv['KEYEVENT']['COD'] in ('USE',) and not aEnv['KEYEVENT']['MOD']:
            # ..............................u (set to user default)
            aEnv['UCS'] = ((1.0, 0.0, aEnv['BORDERCOORD'][0]),\
                           (0.0, -1.0, aEnv['CONTROLLER'].CurrentPage.Height - aEnv['CONTROLLER'].CurrentPage.BorderBottom),\
                           (0.0, 0.0, 1.0))
            aEnv['COORDS'] = [[{'X': aEnv['UCS'][0][2], 'Y': aEnv['UCS'][1][2]}, {'X': aEnv['UCS'][0][2] + 250, 'Y': aEnv['UCS'][1][2]}],\
                       [{'X': aEnv['UCS'][0][2], 'Y': aEnv['UCS'][1][2]}, {'X': aEnv['UCS'][0][2], 'Y': aEnv['UCS'][1][2] - 250}]]
            aEnv['STEP'] = 2
            aEnv['GROUPSHAPE'].remove(aEnv['SHAPE'])
            aEnv['SHAPE'] = aEnv['UCSSHAPE']
            aPosSizeParm = aEnv.fromkeys(aEnv['POSSIZEPARM'])
            for aKey in aPosSizeParm.iterkeys():
                aPosSizeParm[aKey] = aEnv[aKey]
            aEnv.update(self.xEntityPosSize.execute(aPosSizeParm))
            aEnv['SHAPE'] = aEnv['STANDARDSTYLE']
            aEnv['SHAPE'].LineColor = aEnv['STANDARDSTYLE'].LineColor
            aEnv['SHAPE'].LineWidth = aEnv['STANDARDSTYLE'].LineWidth
            aEnv['SHAPE'].setName("ucs")
            aEnv['STOPJOB'] = True
        # --------------------------------------------------------------------------------
        elif aEnv['KEYEVENT']['COD'] in ('ESC',) and not aEnv['KEYEVENT']['MOD']:
            #                                 Esc => kill job
            aEnv['CANCELJOB'] = True
        return aEnv